/****************************************************************************
 *
 * Copyright 2016 Samsung Electronics All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific
 * language governing permissions and limitations under the License.
 *
 ****************************************************************************/
/* address.h -- representation of network addresses
 *
 * Copyright (C) 2010,2011 Olaf Bergmann <bergmann@tzi.org>
 *
 * This file is part of the CoAP library libcoap. Please see
 * README for terms of use.
 */

/**
 * @file protocols/libcoap/address.h
 * @brief representation of network addresses
 */

#ifndef _COAP_ADDRESS_H_
#define _COAP_ADDRESS_H_

#include <protocols/libcoap/config.h>

#ifdef HAVE_ASSERT_H
#include <assert.h>
#else
#ifndef assert
#warning "assertions are disabled"
#define assert(x)
#endif
#endif

#include <string.h>
#include <stdint.h>

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <sys/socket.h>
#endif

#ifdef WITH_LWIP
#include <net/lwip/ipv4/ip_addr.h>

typedef struct coap_address_t {
	uint16_t port;
	ip_addr_t addr;
} coap_address_t;

/* FIXME oversimplification: just assuming it's an ipv4 address instead of
 * looking up the appropraite lwip function */

#define _coap_address_equals_impl(A, B) ((A)->addr.addr == (B)->addr.addr && A->port == B->port)

/* FIXME sure there is something in lwip */

#define _coap_is_mcast_impl(Address) 0

#endif							/* WITH_LWIP */
#ifdef WITH_CONTIKI
#include "uip.h"

typedef struct coap_address_t {
	unsigned char size;
	uip_ipaddr_t addr;
	unsigned short port;
} coap_address_t;

#define _coap_address_equals_impl(A,B)				\
  ((A)->size == (B)->size					\
   && (A)->port == (B)->port					\
   && uip_ipaddr_cmp(&((A)->addr),&((B)->addr)))

#define _coap_is_mcast_impl(Address) uip_is_addr_mcast(&((Address)->addr))
#endif							/* WITH_CONTIKI */
#ifdef WITH_POSIX

/** multi-purpose address abstraction */
typedef struct coap_address_t {
	socklen_t size;		/**< size of addr */
	union {
		struct sockaddr sa;
		struct sockaddr_storage st;
		struct sockaddr_in sin;
		struct sockaddr_in6 sin6;
	} addr;
} coap_address_t;

static inline int _coap_address_equals_impl(const coap_address_t *a, const coap_address_t *b)
{
	if (a->size != b->size || a->addr.sa.sa_family != b->addr.sa.sa_family) {
		return 0;
	}

	/* need to compare only relevant parts of sockaddr_in6 */
	switch (a->addr.sa.sa_family) {
	case AF_INET:
		return a->addr.sin.sin_port == b->addr.sin.sin_port && memcmp(&a->addr.sin.sin_addr, &b->addr.sin.sin_addr, sizeof(struct in_addr)) == 0;
	case AF_INET6:
		return a->addr.sin6.sin6_port == b->addr.sin6.sin6_port && memcmp(&a->addr.sin6.sin6_addr, &b->addr.sin6.sin6_addr, sizeof(struct in6_addr)) == 0;
	default:					/* fall through and signal error */
		;
	}
	return 0;
}

static inline int _coap_is_mcast_impl(const coap_address_t *a)
{
	if (!a) {
		return 0;
	}

	switch (a->addr.sa.sa_family) {
	case AF_INET:
		return IN_MULTICAST(a->addr.sin.sin_addr.s_addr);
	case AF_INET6:
#ifdef IN6_IS_ADDR_MULTICAST
		return IN6_IS_ADDR_MULTICAST(&a->addr.sin6.sin6_addr);
#else
		return 0;
#endif
	default:					/* fall through and signal error */
		;
	}
	return 0;
}
#endif							/* WITH_POSIX */

/**
 * Resets the given coap_address_t object @p addr to its default
 * values.  In particular, the member size must be initialized to the
 * available size for storing addresses.
 *
 * @param addr The coap_address_t object to initialize.
 */
static inline void coap_address_init(coap_address_t *addr)
{
	assert(addr);
	memset(addr, 0, sizeof(coap_address_t));
#ifndef WITH_LWIP
	/* lwip has constandt address sizes and doesn't need the .size part */
	addr->size = sizeof(addr->addr);
#endif
}

/**
 * Compares given address objects @p a and @p b. This function returns
 * @c 1 if addresses are equal, @c 0 otherwise. The parameters @p a
 * and @p b must not be @c NULL;
 */
static inline int coap_address_equals(const coap_address_t *a, const coap_address_t *b)
{
	assert(a);
	assert(b);
	return _coap_address_equals_impl(a, b);
}

/**
 * Checks if given address @p a denotes a multicast address.  This
 * function returns @c 1 if @p a is multicast, @c 0 otherwise.
 */
static inline int coap_is_mcast(const coap_address_t *a)
{
	return a && _coap_is_mcast_impl(a);
}

#endif							/* _COAP_ADDRESS_H_ */
