/****************************************************************************
 *
 * Copyright 2018 Samsung Electronics All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific
 * language governing permissions and limitations under the License.
 *
 ****************************************************************************/
//===----------------------------------------------------------------------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is dual licensed under the MIT and the University of Illinois Open
// Source Licenses. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//

// <string>

// basic_string<charT,traits,Allocator>&
//   replace(size_type pos1, size_type n1, const basic_string<charT,traits,Allocator>& str,
//           size_type pos2, size_type n2=npos);
//  the "=npos" was added in C++14

#include <string>
#include <stdexcept>
#include <algorithm>
#include <cassert>
#include "libcxx_tc_common.h"

#include "test_macros.h"

template <class S>
static int
test(S s, typename S::size_type pos1, typename S::size_type n1,
     S str, typename S::size_type pos2, typename S::size_type n2,
     S expected)
{
    const typename S::size_type old_size = s.size();
    S s0 = s;
    if (pos1 <= old_size && pos2 <= str.size())
    {
        s.replace(pos1, n1, str, pos2, n2);
        LIBCPP_ASSERT(s.__invariants());
        TC_ASSERT_EXPR(s == expected);
        typename S::size_type xlen = std::min(n1, old_size - pos1);
        typename S::size_type rlen = std::min(n2, str.size() - pos2);
        TC_ASSERT_EXPR(s.size() == old_size - xlen + rlen);
    }
#ifndef TEST_HAS_NO_EXCEPTIONS
    else
    {
        try
        {
            s.replace(pos1, n1, str, pos2, n2);
            TC_ASSERT_EXPR(false);
        }
        catch (std::out_of_range&)
        {
            TC_ASSERT_EXPR(pos1 > old_size || pos2 > str.size());
            TC_ASSERT_EXPR(s == s0);
        }
    }
#endif
    return 0;
}

template <class S>
static int
test_npos(S s, typename S::size_type pos1, typename S::size_type n1,
     S str, typename S::size_type pos2,
     S expected)
{
    const typename S::size_type old_size = s.size();
    S s0 = s;
    if (pos1 <= old_size && pos2 <= str.size())
    {
        s.replace(pos1, n1, str, pos2);
        LIBCPP_ASSERT(s.__invariants());
        TC_ASSERT_EXPR(s == expected);
        typename S::size_type xlen = std::min(n1, old_size - pos1);
        typename S::size_type rlen = std::min(S::npos, str.size() - pos2);
        TC_ASSERT_EXPR(s.size() == old_size - xlen + rlen);
    }
#ifndef TEST_HAS_NO_EXCEPTIONS
    else
    {
        try
        {
            s.replace(pos1, n1, str, pos2);
            TC_ASSERT_EXPR(false);
        }
        catch (std::out_of_range&)
        {
            TC_ASSERT_EXPR(pos1 > old_size || pos2 > str.size());
            TC_ASSERT_EXPR(s == s0);
        }
    }
#endif
    return 0;
}


template <class S>
static int test0()
{
    TC_ASSERT_FUNC((test(S(""), 0, 0, S(""), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S(""), 0, 1, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 0, 2, S("12"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 0, 4, S("1234"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 0, 6, S("12345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 1, 2, S("23"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 1, 3, S("234"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 1, 5, S("2345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 2, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 2, 1, S("3"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 2, 2, S("34"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 2, 3, S("345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 2, 4, S("345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 4, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 4, 1, S("5"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 4, 2, S("5"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 5, 1, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 0, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 0, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 1, 8, S("23456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 1, 10, S("234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 5, 1, S("6"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 5, 2, S("67"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 5, 4, S("6789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 5, 5, S("67890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 5, 6, S("67890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 9, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 9, 1, S("0"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 9, 2, S("0"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 10, 1, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 0, 19, S("1234567890123456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 0, 20, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 0, 21, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 1, 18, S("234567890123456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 1, 19, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 1, 20, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 10, 1, S("1"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 10, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 10, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 10, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 10, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 19, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 19, 1, S("0"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 19, 2, S("0"))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 20, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 20, 1, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S(""), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S(""), 0, 1, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 0, 2, S("12"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 0, 4, S("1234"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 0, 6, S("12345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 1, 2, S("23"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 1, 3, S("234"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 1, 5, S("2345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 2, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 2, 1, S("3"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 2, 2, S("34"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 2, 3, S("345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 2, 4, S("345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 4, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 4, 1, S("5"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 4, 2, S("5"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 5, 1, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345"), 6, 0, S("can't happen"))));
    return 0;
}

template <class S>
static int test1()
{
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 0, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 0, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 1, 8, S("23456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 1, 10, S("234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 5, 1, S("6"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 5, 2, S("67"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 5, 4, S("6789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 5, 5, S("67890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 5, 6, S("67890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 9, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 9, 1, S("0"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 9, 2, S("0"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 10, 1, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 0, 19, S("1234567890123456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 0, 20, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 0, 21, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 1, 18, S("234567890123456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 1, 19, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 1, 20, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 10, 1, S("1"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 10, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 10, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 10, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 10, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 19, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 19, 1, S("0"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 19, 2, S("0"))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 20, 0, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 20, 1, S(""))));
    TC_ASSERT_FUNC((test(S(""), 0, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S(""), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S(""), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 0, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 0, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 0, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 0, 6, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 1, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 1, 3, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 1, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 1, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 2, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 2, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 2, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 2, 3, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 2, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 4, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 4, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 4, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 5, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 5, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 0, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 0, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 0, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 0, 11, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 1, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 1, 8, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 1, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 1, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 5, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 5, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 5, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 5, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 5, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 5, 6, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 9, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 9, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 9, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 10, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 10, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 0, 1, S("can't happen"))));
    return 0;
}

template <class S>
static int test2()
{
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 0, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 0, 19, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 0, 20, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 0, 21, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 1, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 1, 18, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 1, 19, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 1, 20, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 10, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 10, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 10, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 10, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 10, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 10, 11, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 19, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 19, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 19, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 20, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 20, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S(""), 1, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S(""), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S(""), 0, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 0, 1, S("1abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 0, 2, S("12abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 0, 4, S("1234abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 0, 5, S("12345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 0, 6, S("12345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 1, 1, S("2abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 1, 2, S("23abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 1, 3, S("234abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 1, 4, S("2345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 1, 5, S("2345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 2, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 2, 1, S("3abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 2, 2, S("34abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 2, 3, S("345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 2, 4, S("345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 4, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 4, 1, S("5abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 4, 2, S("5abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 5, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 0, 1, S("1abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 0, 5, S("12345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 0, 9, S("123456789abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 0, 10, S("1234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 0, 11, S("1234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 1, 1, S("2abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 1, 4, S("2345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 1, 8, S("23456789abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 1, 9, S("234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 1, 10, S("234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 5, 1, S("6abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 5, 2, S("67abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 5, 4, S("6789abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 5, 5, S("67890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 5, 6, S("67890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 9, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 9, 1, S("0abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 9, 2, S("0abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 10, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 0, 1, S("1abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 0, 10, S("1234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 0, 19, S("1234567890123456789abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 0, 20, S("12345678901234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 0, 21, S("12345678901234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 1, 1, S("2abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 1, 9, S("234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 1, 18, S("234567890123456789abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 1, 19, S("2345678901234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 1, 20, S("2345678901234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 10, 1, S("1abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 10, 5, S("12345abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 10, 9, S("123456789abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 10, 10, S("1234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 10, 11, S("1234567890abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 19, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 19, 1, S("0abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 19, 2, S("0abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 20, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 20, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S(""), 0, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S(""), 0, 1, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 0, 0, S("bcde"))));
    return 0;
}

template <class S>
static int test3()
{
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 0, 1, S("1bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 0, 2, S("12bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 0, 4, S("1234bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 0, 5, S("12345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 0, 6, S("12345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 1, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 1, 1, S("2bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 1, 2, S("23bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 1, 3, S("234bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 1, 4, S("2345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 1, 5, S("2345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 2, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 2, 1, S("3bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 2, 2, S("34bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 2, 3, S("345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 2, 4, S("345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 4, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 4, 1, S("5bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 4, 2, S("5bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 5, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 5, 1, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 0, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 0, 1, S("1bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 0, 5, S("12345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 0, 9, S("123456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 0, 10, S("1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 0, 11, S("1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 1, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 1, 1, S("2bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 1, 4, S("2345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 1, 8, S("23456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 1, 9, S("234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 1, 10, S("234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 5, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 5, 1, S("6bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 5, 2, S("67bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 5, 4, S("6789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 5, 5, S("67890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 5, 6, S("67890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 9, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 9, 1, S("0bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 9, 2, S("0bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 10, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 10, 1, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 0, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 0, 1, S("1bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 0, 10, S("1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 0, 19, S("1234567890123456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 0, 20, S("12345678901234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 0, 21, S("12345678901234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 1, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 1, 1, S("2bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 1, 9, S("234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 1, 18, S("234567890123456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 1, 19, S("2345678901234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 1, 20, S("2345678901234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 10, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 10, 1, S("1bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 10, 5, S("12345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 10, 9, S("123456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 10, 10, S("1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 10, 11, S("1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 19, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 19, 1, S("0bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 19, 2, S("0bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 20, 0, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 20, 1, S("bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S(""), 0, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S(""), 0, 1, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 0, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 0, 1, S("1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 0, 2, S("12cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 0, 4, S("1234cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 0, 5, S("12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 0, 6, S("12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 1, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 1, 1, S("2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 1, 2, S("23cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 1, 3, S("234cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 1, 4, S("2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 1, 5, S("2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 2, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 2, 1, S("3cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 2, 2, S("34cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 2, 3, S("345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 2, 4, S("345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 4, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 4, 1, S("5cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 4, 2, S("5cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 5, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 5, 1, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 0, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 0, 1, S("1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 0, 5, S("12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 0, 9, S("123456789cde"))));
    return 0;
}

template <class S>
static int test4()
{
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 0, 10, S("1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 0, 11, S("1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 1, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 1, 1, S("2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 1, 4, S("2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 1, 8, S("23456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 1, 9, S("234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 1, 10, S("234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 5, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 5, 1, S("6cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 5, 2, S("67cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 5, 4, S("6789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 5, 5, S("67890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 5, 6, S("67890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 9, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 9, 1, S("0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 9, 2, S("0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 10, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 10, 1, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 0, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 0, 1, S("1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 0, 10, S("1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 0, 19, S("1234567890123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 0, 20, S("12345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 0, 21, S("12345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 1, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 1, 1, S("2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 1, 9, S("234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 1, 18, S("234567890123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 1, 19, S("2345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 1, 20, S("2345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 10, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 10, 1, S("1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 10, 5, S("12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 10, 9, S("123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 10, 10, S("1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 10, 11, S("1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 19, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 19, 1, S("0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 19, 2, S("0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 20, 0, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 20, 1, S("cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 2, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S(""), 0, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S(""), 0, 1, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 0, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 0, 1, S("1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 0, 2, S("12e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 0, 4, S("1234e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 0, 5, S("12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 0, 6, S("12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 1, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 1, 1, S("2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 1, 2, S("23e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 1, 3, S("234e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 1, 4, S("2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 1, 5, S("2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 2, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 2, 1, S("3e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 2, 2, S("34e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 2, 3, S("345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 2, 4, S("345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 4, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 4, 1, S("5e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 4, 2, S("5e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 5, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 5, 1, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 0, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 0, 1, S("1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 0, 5, S("12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 0, 9, S("123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 0, 10, S("1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 0, 11, S("1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 1, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 1, 1, S("2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 1, 4, S("2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 1, 8, S("23456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 1, 9, S("234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 1, 10, S("234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 5, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 5, 1, S("6e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 5, 2, S("67e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 5, 4, S("6789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 5, 5, S("67890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 5, 6, S("67890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 9, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 9, 1, S("0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 9, 2, S("0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 10, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 10, 1, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 0, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 0, 1, S("1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 0, 10, S("1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 0, 19, S("1234567890123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 0, 20, S("12345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 0, 21, S("12345678901234567890e"))));
    return 0;
}

template <class S>
static int test5()
{
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 1, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 1, 1, S("2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 1, 9, S("234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 1, 18, S("234567890123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 1, 19, S("2345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 1, 20, S("2345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 10, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 10, 1, S("1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 10, 5, S("12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 10, 9, S("123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 10, 10, S("1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 10, 11, S("1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 19, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 19, 1, S("0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 19, 2, S("0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 20, 0, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 20, 1, S("e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 4, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S(""), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S(""), 0, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 0, 2, S("12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 0, 4, S("1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 0, 6, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 1, 2, S("23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 1, 3, S("234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 1, 5, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 2, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 2, 1, S("3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 2, 2, S("34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 2, 3, S("345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 2, 4, S("345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 4, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 4, 1, S("5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 4, 2, S("5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 5, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 0, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 0, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 1, 8, S("23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 1, 10, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 5, 1, S("6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 5, 2, S("67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 5, 4, S("6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 5, 5, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 5, 6, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 9, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 9, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 9, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 10, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 0, 19, S("1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 0, 20, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 0, 21, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 1, 18, S("234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 1, 19, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 1, 20, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 10, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 10, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 10, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 10, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 10, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 19, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 19, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 19, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 20, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 20, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 5, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S(""), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S(""), 0, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 0, 2, S("12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 0, 4, S("1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 0, 5, S("12345"))));
    return 0;
}

template <class S>
static int test6()
{
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 0, 6, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 1, 2, S("23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 1, 3, S("234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 1, 5, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 2, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 2, 1, S("3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 2, 2, S("34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 2, 3, S("345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 2, 4, S("345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 4, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 4, 1, S("5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 4, 2, S("5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 5, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 0, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 0, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 1, 8, S("23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 1, 10, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 5, 1, S("6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 5, 2, S("67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 5, 4, S("6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 5, 5, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 5, 6, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 9, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 9, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 9, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 10, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 0, 19, S("1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 0, 20, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 0, 21, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 1, 18, S("234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 1, 19, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 1, 20, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 10, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 10, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 10, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 10, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 10, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 19, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 19, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 19, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 20, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 20, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcde"), 0, 6, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S(""), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S(""), 0, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 0, 1, S("a1bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 0, 2, S("a12bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 0, 4, S("a1234bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 0, 5, S("a12345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 0, 6, S("a12345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 1, 1, S("a2bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 1, 2, S("a23bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 1, 3, S("a234bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 1, 4, S("a2345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 1, 5, S("a2345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 2, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 2, 1, S("a3bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 2, 2, S("a34bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 2, 3, S("a345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 2, 4, S("a345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 4, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 4, 1, S("a5bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 4, 2, S("a5bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 5, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 0, 1, S("a1bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 0, 5, S("a12345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 0, 9, S("a123456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 0, 10, S("a1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 0, 11, S("a1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 1, 1, S("a2bcde"))));
    return 0;
}

template <class S>
static int test7()
{
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 1, 4, S("a2345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 1, 8, S("a23456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 1, 9, S("a234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 1, 10, S("a234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 5, 1, S("a6bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 5, 2, S("a67bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 5, 4, S("a6789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 5, 5, S("a67890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 5, 6, S("a67890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 9, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 9, 1, S("a0bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 9, 2, S("a0bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 10, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 0, 1, S("a1bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 0, 10, S("a1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 0, 19, S("a1234567890123456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 0, 20, S("a12345678901234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 0, 21, S("a12345678901234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 1, 1, S("a2bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 1, 9, S("a234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 1, 18, S("a234567890123456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 1, 19, S("a2345678901234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 1, 20, S("a2345678901234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 10, 1, S("a1bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 10, 5, S("a12345bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 10, 9, S("a123456789bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 10, 10, S("a1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 10, 11, S("a1234567890bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 19, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 19, 1, S("a0bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 19, 2, S("a0bcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 20, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 20, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S(""), 0, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S(""), 0, 1, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 0, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 0, 1, S("a1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 0, 2, S("a12cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 0, 4, S("a1234cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 0, 5, S("a12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 0, 6, S("a12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 1, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 1, 1, S("a2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 1, 2, S("a23cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 1, 3, S("a234cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 1, 4, S("a2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 1, 5, S("a2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 2, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 2, 1, S("a3cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 2, 2, S("a34cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 2, 3, S("a345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 2, 4, S("a345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 4, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 4, 1, S("a5cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 4, 2, S("a5cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 5, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 5, 1, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 0, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 0, 1, S("a1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 0, 5, S("a12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 0, 9, S("a123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 0, 10, S("a1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 0, 11, S("a1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 1, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 1, 1, S("a2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 1, 4, S("a2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 1, 8, S("a23456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 1, 9, S("a234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 1, 10, S("a234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 5, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 5, 1, S("a6cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 5, 2, S("a67cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 5, 4, S("a6789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 5, 5, S("a67890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 5, 6, S("a67890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 9, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 9, 1, S("a0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 9, 2, S("a0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 10, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 10, 1, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 0, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 0, 1, S("a1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 0, 10, S("a1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 0, 19, S("a1234567890123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 0, 20, S("a12345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 0, 21, S("a12345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 1, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 1, 1, S("a2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 1, 9, S("a234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 1, 18, S("a234567890123456789cde"))));
    return 0;
}

template <class S>
static int test8()
{
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 1, 19, S("a2345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 1, 20, S("a2345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 10, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 10, 1, S("a1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 10, 5, S("a12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 10, 9, S("a123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 10, 10, S("a1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 10, 11, S("a1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 19, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 19, 1, S("a0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 19, 2, S("a0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 20, 0, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 20, 1, S("acde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S(""), 0, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S(""), 0, 1, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 0, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 0, 1, S("a1de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 0, 2, S("a12de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 0, 4, S("a1234de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 0, 5, S("a12345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 0, 6, S("a12345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 1, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 1, 1, S("a2de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 1, 2, S("a23de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 1, 3, S("a234de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 1, 4, S("a2345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 1, 5, S("a2345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 2, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 2, 1, S("a3de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 2, 2, S("a34de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 2, 3, S("a345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 2, 4, S("a345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 4, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 4, 1, S("a5de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 4, 2, S("a5de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 5, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 5, 1, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 0, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 0, 1, S("a1de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 0, 5, S("a12345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 0, 9, S("a123456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 0, 10, S("a1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 0, 11, S("a1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 1, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 1, 1, S("a2de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 1, 4, S("a2345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 1, 8, S("a23456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 1, 9, S("a234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 1, 10, S("a234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 5, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 5, 1, S("a6de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 5, 2, S("a67de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 5, 4, S("a6789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 5, 5, S("a67890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 5, 6, S("a67890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 9, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 9, 1, S("a0de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 9, 2, S("a0de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 10, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 10, 1, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 0, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 0, 1, S("a1de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 0, 10, S("a1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 0, 19, S("a1234567890123456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 0, 20, S("a12345678901234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 0, 21, S("a12345678901234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 1, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 1, 1, S("a2de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 1, 9, S("a234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 1, 18, S("a234567890123456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 1, 19, S("a2345678901234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 1, 20, S("a2345678901234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 10, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 10, 1, S("a1de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 10, 5, S("a12345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 10, 9, S("a123456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 10, 10, S("a1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 10, 11, S("a1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 19, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 19, 1, S("a0de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 19, 2, S("a0de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 20, 0, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 20, 1, S("ade"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 2, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S(""), 0, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S(""), 0, 1, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 0, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 0, 1, S("a1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 0, 2, S("a12e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 0, 4, S("a1234e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 0, 5, S("a12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 0, 6, S("a12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 1, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 1, 1, S("a2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 1, 2, S("a23e"))));
    return 0;
}

template <class S>
static int test9()
{
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 1, 3, S("a234e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 1, 4, S("a2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 1, 5, S("a2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 2, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 2, 1, S("a3e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 2, 2, S("a34e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 2, 3, S("a345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 2, 4, S("a345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 4, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 4, 1, S("a5e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 4, 2, S("a5e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 5, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 5, 1, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 0, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 0, 1, S("a1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 0, 5, S("a12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 0, 9, S("a123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 0, 10, S("a1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 0, 11, S("a1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 1, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 1, 1, S("a2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 1, 4, S("a2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 1, 8, S("a23456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 1, 9, S("a234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 1, 10, S("a234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 5, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 5, 1, S("a6e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 5, 2, S("a67e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 5, 4, S("a6789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 5, 5, S("a67890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 5, 6, S("a67890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 9, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 9, 1, S("a0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 9, 2, S("a0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 10, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 10, 1, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 0, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 0, 1, S("a1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 0, 10, S("a1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 0, 19, S("a1234567890123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 0, 20, S("a12345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 0, 21, S("a12345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 1, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 1, 1, S("a2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 1, 9, S("a234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 1, 18, S("a234567890123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 1, 19, S("a2345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 1, 20, S("a2345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 10, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 10, 1, S("a1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 10, 5, S("a12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 10, 9, S("a123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 10, 10, S("a1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 10, 11, S("a1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 19, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 19, 1, S("a0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 19, 2, S("a0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 20, 0, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 20, 1, S("ae"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 3, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S(""), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S(""), 0, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 0, 2, S("a12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 0, 4, S("a1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 0, 6, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 1, 2, S("a23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 1, 3, S("a234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 1, 5, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 2, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 2, 1, S("a3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 2, 2, S("a34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 2, 3, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 2, 4, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 4, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 4, 1, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 4, 2, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 5, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 0, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 0, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 1, 8, S("a23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 1, 10, S("a234567890"))));
    return 0;
}

template <class S>
static int test10()
{
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 5, 1, S("a6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 5, 2, S("a67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 5, 4, S("a6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 5, 5, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 5, 6, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 9, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 9, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 9, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 10, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 0, 19, S("a1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 0, 20, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 0, 21, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 1, 18, S("a234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 1, 19, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 1, 20, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 10, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 10, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 10, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 10, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 10, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 19, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 19, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 19, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 20, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 20, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 4, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S(""), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S(""), 0, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 0, 2, S("a12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 0, 4, S("a1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 0, 6, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 1, 2, S("a23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 1, 3, S("a234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 1, 5, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 2, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 2, 1, S("a3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 2, 2, S("a34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 2, 3, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 2, 4, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 4, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 4, 1, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 4, 2, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 5, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 0, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 0, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 1, 8, S("a23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 1, 10, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 5, 1, S("a6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 5, 2, S("a67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 5, 4, S("a6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 5, 5, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 5, 6, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 9, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 9, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 9, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 10, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 0, 19, S("a1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 0, 20, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 0, 21, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 1, 18, S("a234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 1, 19, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 1, 20, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 10, 1, S("a1"))));
    return 0;
}

template <class S>
static int test11()
{
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 10, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 10, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 10, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 10, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 19, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 19, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 19, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 20, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 20, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcde"), 1, 5, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S(""), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S(""), 0, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 0, 1, S("ab1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 0, 2, S("ab12cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 0, 4, S("ab1234cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 0, 5, S("ab12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 0, 6, S("ab12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 1, 1, S("ab2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 1, 2, S("ab23cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 1, 3, S("ab234cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 1, 4, S("ab2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 1, 5, S("ab2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 2, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 2, 1, S("ab3cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 2, 2, S("ab34cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 2, 3, S("ab345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 2, 4, S("ab345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 4, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 4, 1, S("ab5cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 4, 2, S("ab5cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 5, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 0, 1, S("ab1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 0, 5, S("ab12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 0, 9, S("ab123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 0, 10, S("ab1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 0, 11, S("ab1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 1, 1, S("ab2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 1, 4, S("ab2345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 1, 8, S("ab23456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 1, 9, S("ab234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 1, 10, S("ab234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 5, 1, S("ab6cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 5, 2, S("ab67cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 5, 4, S("ab6789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 5, 5, S("ab67890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 5, 6, S("ab67890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 9, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 9, 1, S("ab0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 9, 2, S("ab0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 10, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 0, 1, S("ab1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 0, 10, S("ab1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 0, 19, S("ab1234567890123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 0, 20, S("ab12345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 0, 21, S("ab12345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 1, 1, S("ab2cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 1, 9, S("ab234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 1, 18, S("ab234567890123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 1, 19, S("ab2345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 1, 20, S("ab2345678901234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 10, 1, S("ab1cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 10, 5, S("ab12345cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 10, 9, S("ab123456789cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 10, 10, S("ab1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 10, 11, S("ab1234567890cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 19, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 19, 1, S("ab0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 19, 2, S("ab0cde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 20, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 20, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S(""), 0, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S(""), 0, 1, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 0, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 0, 1, S("ab1de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 0, 2, S("ab12de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 0, 4, S("ab1234de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 0, 5, S("ab12345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 0, 6, S("ab12345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 1, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 1, 1, S("ab2de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 1, 2, S("ab23de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 1, 3, S("ab234de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 1, 4, S("ab2345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 1, 5, S("ab2345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 2, 0, S("abde"))));
    return 0;
}

template <class S>
static int test12()
{
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 2, 1, S("ab3de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 2, 2, S("ab34de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 2, 3, S("ab345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 2, 4, S("ab345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 4, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 4, 1, S("ab5de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 4, 2, S("ab5de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 5, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 5, 1, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 0, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 0, 1, S("ab1de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 0, 5, S("ab12345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 0, 9, S("ab123456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 0, 10, S("ab1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 0, 11, S("ab1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 1, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 1, 1, S("ab2de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 1, 4, S("ab2345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 1, 8, S("ab23456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 1, 9, S("ab234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 1, 10, S("ab234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 5, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 5, 1, S("ab6de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 5, 2, S("ab67de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 5, 4, S("ab6789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 5, 5, S("ab67890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 5, 6, S("ab67890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 9, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 9, 1, S("ab0de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 9, 2, S("ab0de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 10, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 10, 1, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 0, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 0, 1, S("ab1de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 0, 10, S("ab1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 0, 19, S("ab1234567890123456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 0, 20, S("ab12345678901234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 0, 21, S("ab12345678901234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 1, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 1, 1, S("ab2de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 1, 9, S("ab234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 1, 18, S("ab234567890123456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 1, 19, S("ab2345678901234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 1, 20, S("ab2345678901234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 10, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 10, 1, S("ab1de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 10, 5, S("ab12345de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 10, 9, S("ab123456789de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 10, 10, S("ab1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 10, 11, S("ab1234567890de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 19, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 19, 1, S("ab0de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 19, 2, S("ab0de"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 20, 0, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 20, 1, S("abde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S(""), 0, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S(""), 0, 1, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 0, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 0, 1, S("ab1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 0, 2, S("ab12e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 0, 4, S("ab1234e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 0, 5, S("ab12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 0, 6, S("ab12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 1, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 1, 1, S("ab2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 1, 2, S("ab23e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 1, 3, S("ab234e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 1, 4, S("ab2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 1, 5, S("ab2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 2, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 2, 1, S("ab3e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 2, 2, S("ab34e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 2, 3, S("ab345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 2, 4, S("ab345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 4, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 4, 1, S("ab5e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 4, 2, S("ab5e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 5, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 5, 1, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 0, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 0, 1, S("ab1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 0, 5, S("ab12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 0, 9, S("ab123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 0, 10, S("ab1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 0, 11, S("ab1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 1, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 1, 1, S("ab2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 1, 4, S("ab2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 1, 8, S("ab23456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 1, 9, S("ab234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 1, 10, S("ab234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 5, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 5, 1, S("ab6e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 5, 2, S("ab67e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 5, 4, S("ab6789e"))));
    return 0;
}

template <class S>
static int test13()
{
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 5, 5, S("ab67890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 5, 6, S("ab67890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 9, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 9, 1, S("ab0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 9, 2, S("ab0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 10, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 10, 1, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 0, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 0, 1, S("ab1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 0, 10, S("ab1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 0, 19, S("ab1234567890123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 0, 20, S("ab12345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 0, 21, S("ab12345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 1, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 1, 1, S("ab2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 1, 9, S("ab234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 1, 18, S("ab234567890123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 1, 19, S("ab2345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 1, 20, S("ab2345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 10, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 10, 1, S("ab1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 10, 5, S("ab12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 10, 9, S("ab123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 10, 10, S("ab1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 10, 11, S("ab1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 19, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 19, 1, S("ab0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 19, 2, S("ab0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 20, 0, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 20, 1, S("abe"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 2, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S(""), 0, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S(""), 0, 1, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 0, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 0, 1, S("ab1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 0, 2, S("ab12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 0, 4, S("ab1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 0, 5, S("ab12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 0, 6, S("ab12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 1, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 1, 1, S("ab2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 1, 2, S("ab23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 1, 3, S("ab234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 1, 4, S("ab2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 1, 5, S("ab2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 2, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 2, 1, S("ab3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 2, 2, S("ab34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 2, 3, S("ab345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 2, 4, S("ab345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 4, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 4, 1, S("ab5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 4, 2, S("ab5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 5, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 5, 1, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 0, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 0, 1, S("ab1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 0, 5, S("ab12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 0, 9, S("ab123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 0, 10, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 0, 11, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 1, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 1, 1, S("ab2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 1, 4, S("ab2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 1, 8, S("ab23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 1, 9, S("ab234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 1, 10, S("ab234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 5, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 5, 1, S("ab6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 5, 2, S("ab67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 5, 4, S("ab6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 5, 5, S("ab67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 5, 6, S("ab67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 9, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 9, 1, S("ab0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 9, 2, S("ab0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 10, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 10, 1, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 0, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 0, 1, S("ab1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 0, 10, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 0, 19, S("ab1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 0, 20, S("ab12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 0, 21, S("ab12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 1, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 1, 1, S("ab2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 1, 9, S("ab234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 1, 18, S("ab234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 1, 19, S("ab2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 1, 20, S("ab2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 10, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 10, 1, S("ab1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 10, 5, S("ab12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 10, 9, S("ab123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 10, 10, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 10, 11, S("ab1234567890"))));
    return 0;
}

template <class S>
static int test14()
{
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 19, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 19, 1, S("ab0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 19, 2, S("ab0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 20, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 20, 1, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 3, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S(""), 0, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S(""), 0, 1, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 0, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 0, 1, S("ab1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 0, 2, S("ab12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 0, 4, S("ab1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 0, 5, S("ab12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 0, 6, S("ab12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 1, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 1, 1, S("ab2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 1, 2, S("ab23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 1, 3, S("ab234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 1, 4, S("ab2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 1, 5, S("ab2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 2, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 2, 1, S("ab3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 2, 2, S("ab34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 2, 3, S("ab345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 2, 4, S("ab345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 4, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 4, 1, S("ab5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 4, 2, S("ab5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 5, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 5, 1, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 0, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 0, 1, S("ab1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 0, 5, S("ab12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 0, 9, S("ab123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 0, 10, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 0, 11, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 1, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 1, 1, S("ab2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 1, 4, S("ab2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 1, 8, S("ab23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 1, 9, S("ab234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 1, 10, S("ab234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 5, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 5, 1, S("ab6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 5, 2, S("ab67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 5, 4, S("ab6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 5, 5, S("ab67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 5, 6, S("ab67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 9, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 9, 1, S("ab0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 9, 2, S("ab0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 10, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 10, 1, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 0, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 0, 1, S("ab1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 0, 10, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 0, 19, S("ab1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 0, 20, S("ab12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 0, 21, S("ab12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 1, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 1, 1, S("ab2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 1, 9, S("ab234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 1, 18, S("ab234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 1, 19, S("ab2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 1, 20, S("ab2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 10, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 10, 1, S("ab1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 10, 5, S("ab12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 10, 9, S("ab123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 10, 10, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 10, 11, S("ab1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 19, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 19, 1, S("ab0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 19, 2, S("ab0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 20, 0, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 20, 1, S("ab"))));
    TC_ASSERT_FUNC((test(S("abcde"), 2, 4, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S(""), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S(""), 0, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 0, 1, S("abcd1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 0, 2, S("abcd12e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 0, 4, S("abcd1234e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 0, 5, S("abcd12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 0, 6, S("abcd12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 1, 1, S("abcd2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 1, 2, S("abcd23e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 1, 3, S("abcd234e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 1, 4, S("abcd2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 1, 5, S("abcd2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 2, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 2, 1, S("abcd3e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 2, 2, S("abcd34e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 2, 3, S("abcd345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 2, 4, S("abcd345e"))));
    return 0;
}

template <class S>
static int test15()
{
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 4, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 4, 1, S("abcd5e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 4, 2, S("abcd5e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 5, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 0, 1, S("abcd1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 0, 5, S("abcd12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 0, 9, S("abcd123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 0, 10, S("abcd1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 0, 11, S("abcd1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 1, 1, S("abcd2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 1, 4, S("abcd2345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 1, 8, S("abcd23456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 1, 9, S("abcd234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 1, 10, S("abcd234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 5, 1, S("abcd6e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 5, 2, S("abcd67e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 5, 4, S("abcd6789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 5, 5, S("abcd67890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 5, 6, S("abcd67890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 9, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 9, 1, S("abcd0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 9, 2, S("abcd0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 10, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 0, 1, S("abcd1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 0, 10, S("abcd1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 0, 19, S("abcd1234567890123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 0, 20, S("abcd12345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 0, 21, S("abcd12345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 1, 1, S("abcd2e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 1, 9, S("abcd234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 1, 18, S("abcd234567890123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 1, 19, S("abcd2345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 1, 20, S("abcd2345678901234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 10, 1, S("abcd1e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 10, 5, S("abcd12345e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 10, 9, S("abcd123456789e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 10, 10, S("abcd1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 10, 11, S("abcd1234567890e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 19, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 19, 1, S("abcd0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 19, 2, S("abcd0e"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 20, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 20, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S(""), 0, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S(""), 0, 1, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 0, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 0, 1, S("abcd1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 0, 2, S("abcd12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 0, 4, S("abcd1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 0, 5, S("abcd12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 0, 6, S("abcd12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 1, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 1, 1, S("abcd2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 1, 2, S("abcd23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 1, 3, S("abcd234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 1, 4, S("abcd2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 1, 5, S("abcd2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 2, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 2, 1, S("abcd3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 2, 2, S("abcd34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 2, 3, S("abcd345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 2, 4, S("abcd345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 4, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 4, 1, S("abcd5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 4, 2, S("abcd5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 5, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 5, 1, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 0, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 0, 1, S("abcd1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 0, 5, S("abcd12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 0, 9, S("abcd123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 0, 10, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 0, 11, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 1, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 1, 1, S("abcd2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 1, 4, S("abcd2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 1, 8, S("abcd23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 1, 9, S("abcd234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 1, 10, S("abcd234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 5, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 5, 1, S("abcd6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 5, 2, S("abcd67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 5, 4, S("abcd6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 5, 5, S("abcd67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 5, 6, S("abcd67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 9, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 9, 1, S("abcd0"))));
    return 0;
}

template <class S>
static int test16()
{
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 9, 2, S("abcd0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 10, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 10, 1, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 0, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 0, 1, S("abcd1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 0, 10, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 0, 19, S("abcd1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 0, 20, S("abcd12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 0, 21, S("abcd12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 1, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 1, 1, S("abcd2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 1, 9, S("abcd234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 1, 18, S("abcd234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 1, 19, S("abcd2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 1, 20, S("abcd2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 10, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 10, 1, S("abcd1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 10, 5, S("abcd12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 10, 9, S("abcd123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 10, 10, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 10, 11, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 19, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 19, 1, S("abcd0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 19, 2, S("abcd0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 20, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 20, 1, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S(""), 0, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S(""), 0, 1, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 0, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 0, 1, S("abcd1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 0, 2, S("abcd12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 0, 4, S("abcd1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 0, 5, S("abcd12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 0, 6, S("abcd12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 1, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 1, 1, S("abcd2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 1, 2, S("abcd23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 1, 3, S("abcd234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 1, 4, S("abcd2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 1, 5, S("abcd2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 2, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 2, 1, S("abcd3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 2, 2, S("abcd34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 2, 3, S("abcd345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 2, 4, S("abcd345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 4, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 4, 1, S("abcd5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 4, 2, S("abcd5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 5, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 5, 1, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 0, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 0, 1, S("abcd1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 0, 5, S("abcd12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 0, 9, S("abcd123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 0, 10, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 0, 11, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 1, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 1, 1, S("abcd2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 1, 4, S("abcd2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 1, 8, S("abcd23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 1, 9, S("abcd234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 1, 10, S("abcd234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 5, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 5, 1, S("abcd6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 5, 2, S("abcd67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 5, 4, S("abcd6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 5, 5, S("abcd67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 5, 6, S("abcd67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 9, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 9, 1, S("abcd0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 9, 2, S("abcd0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 10, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 10, 1, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 0, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 0, 1, S("abcd1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 0, 10, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 0, 19, S("abcd1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 0, 20, S("abcd12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 0, 21, S("abcd12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 1, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 1, 1, S("abcd2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 1, 9, S("abcd234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 1, 18, S("abcd234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 1, 19, S("abcd2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 1, 20, S("abcd2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 10, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 10, 1, S("abcd1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 10, 5, S("abcd12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 10, 9, S("abcd123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 10, 10, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 10, 11, S("abcd1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 19, 0, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 19, 1, S("abcd0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 19, 2, S("abcd0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 20, 0, S("abcd"))));
    return 0;
}

template <class S>
static int test17()
{
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 20, 1, S("abcd"))));
    TC_ASSERT_FUNC((test(S("abcde"), 4, 2, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S(""), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S(""), 0, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 0, 2, S("abcde12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 0, 4, S("abcde1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 0, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 0, 6, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 1, 2, S("abcde23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 1, 3, S("abcde234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 1, 4, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 1, 5, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 2, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 2, 1, S("abcde3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 2, 2, S("abcde34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 2, 3, S("abcde345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 2, 4, S("abcde345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 4, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 4, 1, S("abcde5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 4, 2, S("abcde5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 5, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 0, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 0, 9, S("abcde123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 0, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 0, 11, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 1, 4, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 1, 8, S("abcde23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 1, 9, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 1, 10, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 5, 1, S("abcde6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 5, 2, S("abcde67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 5, 4, S("abcde6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 5, 5, S("abcde67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 5, 6, S("abcde67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 9, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 9, 1, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 9, 2, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 10, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 0, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 0, 19, S("abcde1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 0, 20, S("abcde12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 0, 21, S("abcde12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 1, 9, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 1, 18, S("abcde234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 1, 19, S("abcde2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 1, 20, S("abcde2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 10, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 10, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 10, 9, S("abcde123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 10, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 10, 11, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 19, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 19, 1, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 19, 2, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 20, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 20, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S(""), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S(""), 0, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 0, 2, S("abcde12"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 0, 4, S("abcde1234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 0, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 0, 6, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 1, 2, S("abcde23"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 1, 3, S("abcde234"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 1, 4, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 1, 5, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 2, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 2, 1, S("abcde3"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 2, 2, S("abcde34"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 2, 3, S("abcde345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 2, 4, S("abcde345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 4, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 4, 1, S("abcde5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 4, 2, S("abcde5"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 5, 0, S("abcde"))));
    return 0;
}

template <class S>
static int test18()
{
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 5, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 0, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 0, 9, S("abcde123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 0, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 0, 11, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 1, 4, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 1, 8, S("abcde23456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 1, 9, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 1, 10, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 5, 1, S("abcde6"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 5, 2, S("abcde67"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 5, 4, S("abcde6789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 5, 5, S("abcde67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 5, 6, S("abcde67890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 9, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 9, 1, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 9, 2, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 10, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 0, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 0, 19, S("abcde1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 0, 20, S("abcde12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 0, 21, S("abcde12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 1, 9, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 1, 18, S("abcde234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 1, 19, S("abcde2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 1, 20, S("abcde2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 10, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 10, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 10, 9, S("abcde123456789"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 10, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 10, 11, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 19, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 19, 1, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 19, 2, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 20, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 20, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcde"), 5, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S(""), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S(""), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 0, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 0, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 0, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 0, 6, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 1, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 1, 3, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 1, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 1, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 2, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 2, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 2, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 2, 3, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 2, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 4, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 4, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 4, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 5, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 5, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 0, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 0, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 0, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 0, 11, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 1, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 1, 8, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 1, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 1, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 5, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 5, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 5, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 5, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 5, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 5, 6, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 9, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 9, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 9, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 10, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 10, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("1234567890"), 11, 0, S("can't happen"))));
    return 0;
}

template <class S>
static int test19()
{
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 0, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 0, 19, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 0, 20, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 0, 21, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 1, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 1, 18, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 1, 19, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 1, 20, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 10, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 10, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 10, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 10, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 10, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 10, 11, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 19, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 19, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 19, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 20, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 20, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcde"), 6, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S(""), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S(""), 0, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 0, 1, S("1abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 0, 2, S("12abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 0, 4, S("1234abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 0, 5, S("12345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 0, 6, S("12345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 1, 1, S("2abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 1, 2, S("23abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 1, 3, S("234abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 1, 4, S("2345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 1, 5, S("2345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 2, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 2, 1, S("3abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 2, 2, S("34abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 2, 3, S("345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 2, 4, S("345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 4, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 4, 1, S("5abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 4, 2, S("5abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 5, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 0, 1, S("1abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 0, 5, S("12345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 0, 9, S("123456789abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 0, 10, S("1234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 0, 11, S("1234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 1, 1, S("2abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 1, 4, S("2345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 1, 8, S("23456789abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 1, 9, S("234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 1, 10, S("234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 5, 1, S("6abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 5, 2, S("67abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 5, 4, S("6789abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 5, 5, S("67890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 5, 6, S("67890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 9, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 9, 1, S("0abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 9, 2, S("0abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 10, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 0, 1, S("1abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 0, 10, S("1234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 0, 19, S("1234567890123456789abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 0, 20, S("12345678901234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 0, 21, S("12345678901234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 1, 1, S("2abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 1, 9, S("234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 1, 18, S("234567890123456789abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 1, 19, S("2345678901234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 1, 20, S("2345678901234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 10, 1, S("1abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 10, 5, S("12345abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 10, 9, S("123456789abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 10, 10, S("1234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 10, 11, S("1234567890abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 19, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 19, 1, S("0abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 19, 2, S("0abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 20, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 20, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S(""), 0, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S(""), 0, 1, S("bcdefghij"))));
    return 0;
}

template <class S>
static int test20()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 0, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 0, 1, S("1bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 0, 2, S("12bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 0, 4, S("1234bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 0, 5, S("12345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 0, 6, S("12345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 1, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 1, 1, S("2bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 1, 2, S("23bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 1, 3, S("234bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 1, 4, S("2345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 1, 5, S("2345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 2, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 2, 1, S("3bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 2, 2, S("34bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 2, 3, S("345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 2, 4, S("345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 4, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 4, 1, S("5bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 4, 2, S("5bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 5, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 5, 1, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 0, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 0, 1, S("1bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 0, 5, S("12345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 0, 9, S("123456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 0, 10, S("1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 0, 11, S("1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 1, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 1, 1, S("2bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 1, 4, S("2345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 1, 8, S("23456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 1, 9, S("234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 1, 10, S("234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 5, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 5, 1, S("6bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 5, 2, S("67bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 5, 4, S("6789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 5, 5, S("67890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 5, 6, S("67890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 9, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 9, 1, S("0bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 9, 2, S("0bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 10, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 10, 1, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 0, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 0, 1, S("1bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 0, 10, S("1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 0, 19, S("1234567890123456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 0, 20, S("12345678901234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 0, 21, S("12345678901234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 1, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 1, 1, S("2bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 1, 9, S("234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 1, 18, S("234567890123456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 1, 19, S("2345678901234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 1, 20, S("2345678901234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 10, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 10, 1, S("1bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 10, 5, S("12345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 10, 9, S("123456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 10, 10, S("1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 10, 11, S("1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 19, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 19, 1, S("0bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 19, 2, S("0bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 20, 0, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 20, 1, S("bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S(""), 0, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S(""), 0, 1, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 0, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 0, 1, S("1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 0, 2, S("12fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 0, 4, S("1234fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 0, 5, S("12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 0, 6, S("12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 1, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 1, 1, S("2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 1, 2, S("23fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 1, 3, S("234fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 1, 4, S("2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 1, 5, S("2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 2, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 2, 1, S("3fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 2, 2, S("34fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 2, 3, S("345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 2, 4, S("345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 4, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 4, 1, S("5fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 4, 2, S("5fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 5, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 5, 1, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 0, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 0, 1, S("1fghij"))));
    return 0;
}

template <class S>
static int test21()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 0, 5, S("12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 0, 9, S("123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 0, 10, S("1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 0, 11, S("1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 1, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 1, 1, S("2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 1, 4, S("2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 1, 8, S("23456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 1, 9, S("234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 1, 10, S("234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 5, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 5, 1, S("6fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 5, 2, S("67fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 5, 4, S("6789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 5, 5, S("67890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 5, 6, S("67890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 9, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 9, 1, S("0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 9, 2, S("0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 10, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 10, 1, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 0, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 0, 1, S("1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 0, 10, S("1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 0, 19, S("1234567890123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 0, 20, S("12345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 0, 21, S("12345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 1, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 1, 1, S("2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 1, 9, S("234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 1, 18, S("234567890123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 1, 19, S("2345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 1, 20, S("2345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 10, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 10, 1, S("1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 10, 5, S("12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 10, 9, S("123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 10, 10, S("1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 10, 11, S("1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 19, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 19, 1, S("0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 19, 2, S("0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 20, 0, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 20, 1, S("fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 5, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S(""), 0, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S(""), 0, 1, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 0, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 0, 1, S("1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 0, 2, S("12j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 0, 4, S("1234j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 0, 5, S("12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 0, 6, S("12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 1, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 1, 1, S("2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 1, 2, S("23j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 1, 3, S("234j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 1, 4, S("2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 1, 5, S("2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 2, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 2, 1, S("3j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 2, 2, S("34j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 2, 3, S("345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 2, 4, S("345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 4, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 4, 1, S("5j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 4, 2, S("5j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 5, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 5, 1, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 0, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 0, 1, S("1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 0, 5, S("12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 0, 9, S("123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 0, 10, S("1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 0, 11, S("1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 1, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 1, 1, S("2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 1, 4, S("2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 1, 8, S("23456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 1, 9, S("234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 1, 10, S("234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 5, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 5, 1, S("6j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 5, 2, S("67j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 5, 4, S("6789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 5, 5, S("67890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 5, 6, S("67890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 9, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 9, 1, S("0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 9, 2, S("0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 10, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 10, 1, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 0, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 0, 1, S("1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 0, 10, S("1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 0, 19, S("1234567890123456789j"))));
    return 0;
}

template <class S>
static int test22()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 0, 20, S("12345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 0, 21, S("12345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 1, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 1, 1, S("2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 1, 9, S("234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 1, 18, S("234567890123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 1, 19, S("2345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 1, 20, S("2345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 10, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 10, 1, S("1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 10, 5, S("12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 10, 9, S("123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 10, 10, S("1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 10, 11, S("1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 19, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 19, 1, S("0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 19, 2, S("0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 20, 0, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 20, 1, S("j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 9, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S(""), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S(""), 0, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 0, 2, S("12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 0, 4, S("1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 0, 6, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 1, 2, S("23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 1, 3, S("234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 1, 5, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 2, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 2, 1, S("3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 2, 2, S("34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 2, 3, S("345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 2, 4, S("345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 4, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 4, 1, S("5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 4, 2, S("5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 5, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 0, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 0, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 1, 8, S("23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 1, 10, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 5, 1, S("6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 5, 2, S("67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 5, 4, S("6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 5, 5, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 5, 6, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 9, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 9, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 9, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 10, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 0, 19, S("1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 0, 20, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 0, 21, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 1, 18, S("234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 1, 19, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 1, 20, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 10, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 10, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 10, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 10, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 10, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 19, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 19, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 19, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 20, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 20, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 10, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S(""), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S(""), 0, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 0, 2, S("12"))));
    return 0;
}

template <class S>
static int test23()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 0, 4, S("1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 0, 6, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 1, 2, S("23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 1, 3, S("234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 1, 5, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 2, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 2, 1, S("3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 2, 2, S("34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 2, 3, S("345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 2, 4, S("345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 4, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 4, 1, S("5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 4, 2, S("5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 5, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 0, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 0, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 1, 8, S("23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 1, 10, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 5, 1, S("6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 5, 2, S("67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 5, 4, S("6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 5, 5, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 5, 6, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 9, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 9, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 9, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 10, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 0, 19, S("1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 0, 20, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 0, 21, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 1, 18, S("234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 1, 19, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 1, 20, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 10, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 10, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 10, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 10, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 10, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 19, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 19, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 19, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 20, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 20, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 0, 11, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S(""), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S(""), 0, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 0, 1, S("a1bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 0, 2, S("a12bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 0, 4, S("a1234bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 0, 5, S("a12345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 0, 6, S("a12345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 1, 1, S("a2bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 1, 2, S("a23bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 1, 3, S("a234bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 1, 4, S("a2345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 1, 5, S("a2345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 2, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 2, 1, S("a3bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 2, 2, S("a34bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 2, 3, S("a345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 2, 4, S("a345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 4, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 4, 1, S("a5bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 4, 2, S("a5bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 5, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 0, 1, S("a1bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 0, 5, S("a12345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 0, 9, S("a123456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 0, 10, S("a1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 0, 11, S("a1234567890bcdefghij"))));
    return 0;
}

template <class S>
static int test24()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 1, 1, S("a2bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 1, 4, S("a2345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 1, 8, S("a23456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 1, 9, S("a234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 1, 10, S("a234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 5, 1, S("a6bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 5, 2, S("a67bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 5, 4, S("a6789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 5, 5, S("a67890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 5, 6, S("a67890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 9, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 9, 1, S("a0bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 9, 2, S("a0bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 10, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 0, 1, S("a1bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 0, 10, S("a1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 0, 19, S("a1234567890123456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 0, 20, S("a12345678901234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 0, 21, S("a12345678901234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 1, 1, S("a2bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 1, 9, S("a234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 1, 18, S("a234567890123456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 1, 19, S("a2345678901234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 1, 20, S("a2345678901234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 10, 1, S("a1bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 10, 5, S("a12345bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 10, 9, S("a123456789bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 10, 10, S("a1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 10, 11, S("a1234567890bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 19, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 19, 1, S("a0bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 19, 2, S("a0bcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 20, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 20, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S(""), 0, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S(""), 0, 1, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 0, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 0, 1, S("a1cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 0, 2, S("a12cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 0, 4, S("a1234cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 0, 5, S("a12345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 0, 6, S("a12345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 1, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 1, 1, S("a2cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 1, 2, S("a23cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 1, 3, S("a234cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 1, 4, S("a2345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 1, 5, S("a2345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 2, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 2, 1, S("a3cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 2, 2, S("a34cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 2, 3, S("a345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 2, 4, S("a345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 4, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 4, 1, S("a5cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 4, 2, S("a5cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 5, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 5, 1, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 0, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 0, 1, S("a1cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 0, 5, S("a12345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 0, 9, S("a123456789cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 0, 10, S("a1234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 0, 11, S("a1234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 1, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 1, 1, S("a2cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 1, 4, S("a2345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 1, 8, S("a23456789cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 1, 9, S("a234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 1, 10, S("a234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 5, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 5, 1, S("a6cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 5, 2, S("a67cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 5, 4, S("a6789cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 5, 5, S("a67890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 5, 6, S("a67890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 9, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 9, 1, S("a0cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 9, 2, S("a0cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 10, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 10, 1, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 0, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 0, 1, S("a1cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 0, 10, S("a1234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 0, 19, S("a1234567890123456789cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 0, 20, S("a12345678901234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 0, 21, S("a12345678901234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 1, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 1, 1, S("a2cdefghij"))));
    return 0;
}

template <class S>
static int test25()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 1, 9, S("a234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 1, 18, S("a234567890123456789cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 1, 19, S("a2345678901234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 1, 20, S("a2345678901234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 10, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 10, 1, S("a1cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 10, 5, S("a12345cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 10, 9, S("a123456789cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 10, 10, S("a1234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 10, 11, S("a1234567890cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 19, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 19, 1, S("a0cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 19, 2, S("a0cdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 20, 0, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 20, 1, S("acdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S(""), 0, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S(""), 0, 1, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 0, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 0, 1, S("a1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 0, 2, S("a12fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 0, 4, S("a1234fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 0, 5, S("a12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 0, 6, S("a12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 1, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 1, 1, S("a2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 1, 2, S("a23fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 1, 3, S("a234fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 1, 4, S("a2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 1, 5, S("a2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 2, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 2, 1, S("a3fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 2, 2, S("a34fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 2, 3, S("a345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 2, 4, S("a345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 4, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 4, 1, S("a5fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 4, 2, S("a5fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 5, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 5, 1, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 0, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 0, 1, S("a1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 0, 5, S("a12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 0, 9, S("a123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 0, 10, S("a1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 0, 11, S("a1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 1, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 1, 1, S("a2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 1, 4, S("a2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 1, 8, S("a23456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 1, 9, S("a234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 1, 10, S("a234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 5, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 5, 1, S("a6fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 5, 2, S("a67fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 5, 4, S("a6789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 5, 5, S("a67890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 5, 6, S("a67890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 9, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 9, 1, S("a0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 9, 2, S("a0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 10, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 10, 1, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 0, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 0, 1, S("a1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 0, 10, S("a1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 0, 19, S("a1234567890123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 0, 20, S("a12345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 0, 21, S("a12345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 1, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 1, 1, S("a2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 1, 9, S("a234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 1, 18, S("a234567890123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 1, 19, S("a2345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 1, 20, S("a2345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 10, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 10, 1, S("a1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 10, 5, S("a12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 10, 9, S("a123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 10, 10, S("a1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 10, 11, S("a1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 19, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 19, 1, S("a0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 19, 2, S("a0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 20, 0, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 20, 1, S("afghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 4, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S(""), 0, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S(""), 0, 1, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 0, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 0, 1, S("a1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 0, 2, S("a12j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 0, 4, S("a1234j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 0, 5, S("a12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 0, 6, S("a12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 1, 0, S("aj"))));
    return 0;
}

template <class S>
static int test26()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 1, 1, S("a2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 1, 2, S("a23j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 1, 3, S("a234j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 1, 4, S("a2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 1, 5, S("a2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 2, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 2, 1, S("a3j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 2, 2, S("a34j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 2, 3, S("a345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 2, 4, S("a345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 4, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 4, 1, S("a5j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 4, 2, S("a5j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 5, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 5, 1, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 0, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 0, 1, S("a1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 0, 5, S("a12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 0, 9, S("a123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 0, 10, S("a1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 0, 11, S("a1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 1, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 1, 1, S("a2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 1, 4, S("a2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 1, 8, S("a23456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 1, 9, S("a234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 1, 10, S("a234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 5, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 5, 1, S("a6j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 5, 2, S("a67j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 5, 4, S("a6789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 5, 5, S("a67890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 5, 6, S("a67890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 9, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 9, 1, S("a0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 9, 2, S("a0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 10, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 10, 1, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 0, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 0, 1, S("a1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 0, 10, S("a1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 0, 19, S("a1234567890123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 0, 20, S("a12345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 0, 21, S("a12345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 1, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 1, 1, S("a2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 1, 9, S("a234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 1, 18, S("a234567890123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 1, 19, S("a2345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 1, 20, S("a2345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 10, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 10, 1, S("a1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 10, 5, S("a12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 10, 9, S("a123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 10, 10, S("a1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 10, 11, S("a1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 19, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 19, 1, S("a0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 19, 2, S("a0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 20, 0, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 20, 1, S("aj"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 8, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S(""), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S(""), 0, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 0, 2, S("a12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 0, 4, S("a1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 0, 6, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 1, 2, S("a23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 1, 3, S("a234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 1, 5, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 2, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 2, 1, S("a3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 2, 2, S("a34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 2, 3, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 2, 4, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 4, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 4, 1, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 4, 2, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 5, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 0, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 0, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 1, 8, S("a23456789"))));
    return 0;
}

template <class S>
static int test27()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 1, 10, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 5, 1, S("a6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 5, 2, S("a67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 5, 4, S("a6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 5, 5, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 5, 6, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 9, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 9, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 9, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 10, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 0, 19, S("a1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 0, 20, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 0, 21, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 1, 18, S("a234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 1, 19, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 1, 20, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 10, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 10, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 10, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 10, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 10, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 19, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 19, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 19, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 20, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 20, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 9, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S(""), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S(""), 0, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 0, 2, S("a12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 0, 4, S("a1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 0, 6, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 1, 2, S("a23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 1, 3, S("a234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 1, 5, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 2, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 2, 1, S("a3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 2, 2, S("a34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 2, 3, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 2, 4, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 4, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 4, 1, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 4, 2, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 5, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 0, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 0, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 1, 8, S("a23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 1, 10, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 5, 1, S("a6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 5, 2, S("a67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 5, 4, S("a6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 5, 5, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 5, 6, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 9, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 9, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 9, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 10, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 0, 19, S("a1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 0, 20, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 0, 21, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 1, 18, S("a234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 1, 19, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 1, 20, S("a2345678901234567890"))));
    return 0;
}

template <class S>
static int test28()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 10, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 10, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 10, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 10, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 10, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 19, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 19, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 19, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 20, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 20, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 1, 10, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S(""), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S(""), 0, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 0, 1, S("abcde1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 0, 2, S("abcde12fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 0, 4, S("abcde1234fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 0, 5, S("abcde12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 0, 6, S("abcde12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 1, 1, S("abcde2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 1, 2, S("abcde23fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 1, 3, S("abcde234fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 1, 4, S("abcde2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 1, 5, S("abcde2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 2, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 2, 1, S("abcde3fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 2, 2, S("abcde34fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 2, 3, S("abcde345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 2, 4, S("abcde345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 4, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 4, 1, S("abcde5fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 4, 2, S("abcde5fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 5, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 0, 1, S("abcde1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 0, 5, S("abcde12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 0, 9, S("abcde123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 0, 10, S("abcde1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 0, 11, S("abcde1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 1, 1, S("abcde2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 1, 4, S("abcde2345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 1, 8, S("abcde23456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 1, 9, S("abcde234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 1, 10, S("abcde234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 5, 1, S("abcde6fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 5, 2, S("abcde67fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 5, 4, S("abcde6789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 5, 5, S("abcde67890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 5, 6, S("abcde67890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 9, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 9, 1, S("abcde0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 9, 2, S("abcde0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 10, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 0, 1, S("abcde1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 0, 10, S("abcde1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 0, 19, S("abcde1234567890123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 0, 20, S("abcde12345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 0, 21, S("abcde12345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 1, 1, S("abcde2fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 1, 9, S("abcde234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 1, 18, S("abcde234567890123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 1, 19, S("abcde2345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 1, 20, S("abcde2345678901234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 10, 1, S("abcde1fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 10, 5, S("abcde12345fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 10, 9, S("abcde123456789fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 10, 10, S("abcde1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 10, 11, S("abcde1234567890fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 19, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 19, 1, S("abcde0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 19, 2, S("abcde0fghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 20, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 20, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S(""), 0, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S(""), 0, 1, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 0, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 0, 1, S("abcde1ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 0, 2, S("abcde12ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 0, 4, S("abcde1234ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 0, 5, S("abcde12345ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 0, 6, S("abcde12345ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 1, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 1, 1, S("abcde2ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 1, 2, S("abcde23ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 1, 3, S("abcde234ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 1, 4, S("abcde2345ghij"))));
    return 0;
}

template <class S>
static int test29()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 1, 5, S("abcde2345ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 2, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 2, 1, S("abcde3ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 2, 2, S("abcde34ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 2, 3, S("abcde345ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 2, 4, S("abcde345ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 4, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 4, 1, S("abcde5ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 4, 2, S("abcde5ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 5, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 5, 1, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 0, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 0, 1, S("abcde1ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 0, 5, S("abcde12345ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 0, 9, S("abcde123456789ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 0, 10, S("abcde1234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 0, 11, S("abcde1234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 1, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 1, 1, S("abcde2ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 1, 4, S("abcde2345ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 1, 8, S("abcde23456789ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 1, 9, S("abcde234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 1, 10, S("abcde234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 5, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 5, 1, S("abcde6ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 5, 2, S("abcde67ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 5, 4, S("abcde6789ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 5, 5, S("abcde67890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 5, 6, S("abcde67890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 9, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 9, 1, S("abcde0ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 9, 2, S("abcde0ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 10, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 10, 1, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 0, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 0, 1, S("abcde1ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 0, 10, S("abcde1234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 0, 19, S("abcde1234567890123456789ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 0, 20, S("abcde12345678901234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 0, 21, S("abcde12345678901234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 1, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 1, 1, S("abcde2ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 1, 9, S("abcde234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 1, 18, S("abcde234567890123456789ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 1, 19, S("abcde2345678901234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 1, 20, S("abcde2345678901234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 10, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 10, 1, S("abcde1ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 10, 5, S("abcde12345ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 10, 9, S("abcde123456789ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 10, 10, S("abcde1234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 10, 11, S("abcde1234567890ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 19, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 19, 1, S("abcde0ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 19, 2, S("abcde0ghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 20, 0, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 20, 1, S("abcdeghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S(""), 0, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S(""), 0, 1, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 0, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 0, 1, S("abcde1hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 0, 2, S("abcde12hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 0, 4, S("abcde1234hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 0, 5, S("abcde12345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 0, 6, S("abcde12345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 1, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 1, 1, S("abcde2hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 1, 2, S("abcde23hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 1, 3, S("abcde234hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 1, 4, S("abcde2345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 1, 5, S("abcde2345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 2, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 2, 1, S("abcde3hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 2, 2, S("abcde34hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 2, 3, S("abcde345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 2, 4, S("abcde345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 4, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 4, 1, S("abcde5hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 4, 2, S("abcde5hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 5, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 5, 1, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 0, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 0, 1, S("abcde1hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 0, 5, S("abcde12345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 0, 9, S("abcde123456789hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 0, 10, S("abcde1234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 0, 11, S("abcde1234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 1, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 1, 1, S("abcde2hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 1, 4, S("abcde2345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 1, 8, S("abcde23456789hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 1, 9, S("abcde234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 1, 10, S("abcde234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 5, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 5, 1, S("abcde6hij"))));
    return 0;
}

template <class S>
static int test30()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 5, 2, S("abcde67hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 5, 4, S("abcde6789hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 5, 5, S("abcde67890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 5, 6, S("abcde67890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 9, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 9, 1, S("abcde0hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 9, 2, S("abcde0hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 10, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 10, 1, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 0, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 0, 1, S("abcde1hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 0, 10, S("abcde1234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 0, 19, S("abcde1234567890123456789hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 0, 20, S("abcde12345678901234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 0, 21, S("abcde12345678901234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 1, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 1, 1, S("abcde2hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 1, 9, S("abcde234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 1, 18, S("abcde234567890123456789hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 1, 19, S("abcde2345678901234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 1, 20, S("abcde2345678901234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 10, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 10, 1, S("abcde1hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 10, 5, S("abcde12345hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 10, 9, S("abcde123456789hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 10, 10, S("abcde1234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 10, 11, S("abcde1234567890hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 19, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 19, 1, S("abcde0hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 19, 2, S("abcde0hij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 20, 0, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 20, 1, S("abcdehij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 2, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S(""), 0, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S(""), 0, 1, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 0, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 0, 1, S("abcde1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 0, 2, S("abcde12j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 0, 4, S("abcde1234j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 0, 5, S("abcde12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 0, 6, S("abcde12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 1, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 1, 1, S("abcde2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 1, 2, S("abcde23j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 1, 3, S("abcde234j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 1, 4, S("abcde2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 1, 5, S("abcde2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 2, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 2, 1, S("abcde3j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 2, 2, S("abcde34j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 2, 3, S("abcde345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 2, 4, S("abcde345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 4, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 4, 1, S("abcde5j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 4, 2, S("abcde5j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 5, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 5, 1, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 0, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 0, 1, S("abcde1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 0, 5, S("abcde12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 0, 9, S("abcde123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 0, 10, S("abcde1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 0, 11, S("abcde1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 1, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 1, 1, S("abcde2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 1, 4, S("abcde2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 1, 8, S("abcde23456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 1, 9, S("abcde234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 1, 10, S("abcde234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 5, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 5, 1, S("abcde6j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 5, 2, S("abcde67j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 5, 4, S("abcde6789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 5, 5, S("abcde67890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 5, 6, S("abcde67890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 9, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 9, 1, S("abcde0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 9, 2, S("abcde0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 10, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 10, 1, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 0, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 0, 1, S("abcde1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 0, 10, S("abcde1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 0, 19, S("abcde1234567890123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 0, 20, S("abcde12345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 0, 21, S("abcde12345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 1, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 1, 1, S("abcde2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 1, 9, S("abcde234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 1, 18, S("abcde234567890123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 1, 19, S("abcde2345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 1, 20, S("abcde2345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 10, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 10, 1, S("abcde1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 10, 5, S("abcde12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 10, 9, S("abcde123456789j"))));
    return 0;
}

template <class S>
static int test31()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 10, 10, S("abcde1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 10, 11, S("abcde1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 19, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 19, 1, S("abcde0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 19, 2, S("abcde0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 20, 0, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 20, 1, S("abcdej"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 4, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S(""), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S(""), 0, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 0, 2, S("abcde12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 0, 4, S("abcde1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 0, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 0, 6, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 1, 2, S("abcde23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 1, 3, S("abcde234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 1, 4, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 1, 5, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 2, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 2, 1, S("abcde3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 2, 2, S("abcde34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 2, 3, S("abcde345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 2, 4, S("abcde345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 4, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 4, 1, S("abcde5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 4, 2, S("abcde5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 5, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 0, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 0, 9, S("abcde123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 0, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 0, 11, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 1, 4, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 1, 8, S("abcde23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 1, 9, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 1, 10, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 5, 1, S("abcde6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 5, 2, S("abcde67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 5, 4, S("abcde6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 5, 5, S("abcde67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 5, 6, S("abcde67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 9, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 9, 1, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 9, 2, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 10, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 0, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 0, 19, S("abcde1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 0, 20, S("abcde12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 0, 21, S("abcde12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 1, 9, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 1, 18, S("abcde234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 1, 19, S("abcde2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 1, 20, S("abcde2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 10, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 10, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 10, 9, S("abcde123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 10, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 10, 11, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 19, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 19, 1, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 19, 2, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 20, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 20, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 5, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S(""), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S(""), 0, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 0, 2, S("abcde12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 0, 4, S("abcde1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 0, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 0, 6, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 1, 2, S("abcde23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 1, 3, S("abcde234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 1, 4, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 1, 5, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 2, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 2, 1, S("abcde3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 2, 2, S("abcde34"))));
    return 0;
}

template <class S>
static int test32()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 2, 3, S("abcde345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 2, 4, S("abcde345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 4, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 4, 1, S("abcde5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 4, 2, S("abcde5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 5, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 0, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 0, 9, S("abcde123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 0, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 0, 11, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 1, 4, S("abcde2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 1, 8, S("abcde23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 1, 9, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 1, 10, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 5, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 5, 1, S("abcde6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 5, 2, S("abcde67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 5, 4, S("abcde6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 5, 5, S("abcde67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 5, 6, S("abcde67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 9, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 9, 1, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 9, 2, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 10, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 0, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 0, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 0, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 0, 19, S("abcde1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 0, 20, S("abcde12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 0, 21, S("abcde12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 1, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 1, 1, S("abcde2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 1, 9, S("abcde234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 1, 18, S("abcde234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 1, 19, S("abcde2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 1, 20, S("abcde2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 10, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 10, 1, S("abcde1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 10, 5, S("abcde12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 10, 9, S("abcde123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 10, 10, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 10, 11, S("abcde1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 19, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 19, 1, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 19, 2, S("abcde0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 20, 0, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 20, 1, S("abcde"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 5, 6, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S(""), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S(""), 0, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 0, 1, S("abcdefghi1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 0, 2, S("abcdefghi12j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 0, 4, S("abcdefghi1234j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 0, 5, S("abcdefghi12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 0, 6, S("abcdefghi12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 1, 1, S("abcdefghi2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 1, 2, S("abcdefghi23j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 1, 3, S("abcdefghi234j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 1, 4, S("abcdefghi2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 1, 5, S("abcdefghi2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 2, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 2, 1, S("abcdefghi3j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 2, 2, S("abcdefghi34j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 2, 3, S("abcdefghi345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 2, 4, S("abcdefghi345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 4, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 4, 1, S("abcdefghi5j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 4, 2, S("abcdefghi5j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 5, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 0, 1, S("abcdefghi1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 0, 5, S("abcdefghi12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 0, 9, S("abcdefghi123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 0, 10, S("abcdefghi1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 0, 11, S("abcdefghi1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 1, 1, S("abcdefghi2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 1, 4, S("abcdefghi2345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 1, 8, S("abcdefghi23456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 1, 9, S("abcdefghi234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 1, 10, S("abcdefghi234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 5, 1, S("abcdefghi6j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 5, 2, S("abcdefghi67j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 5, 4, S("abcdefghi6789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 5, 5, S("abcdefghi67890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 5, 6, S("abcdefghi67890j"))));
    return 0;
}

template <class S>
static int test33()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 9, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 9, 1, S("abcdefghi0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 9, 2, S("abcdefghi0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 10, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 0, 1, S("abcdefghi1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 0, 10, S("abcdefghi1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 0, 19, S("abcdefghi1234567890123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 0, 20, S("abcdefghi12345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 0, 21, S("abcdefghi12345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 1, 1, S("abcdefghi2j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 1, 9, S("abcdefghi234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 1, 18, S("abcdefghi234567890123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 1, 19, S("abcdefghi2345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 1, 20, S("abcdefghi2345678901234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 10, 1, S("abcdefghi1j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 10, 5, S("abcdefghi12345j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 10, 9, S("abcdefghi123456789j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 10, 10, S("abcdefghi1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 10, 11, S("abcdefghi1234567890j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 19, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 19, 1, S("abcdefghi0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 19, 2, S("abcdefghi0j"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 20, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 20, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S(""), 0, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S(""), 0, 1, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 0, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 0, 1, S("abcdefghi1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 0, 2, S("abcdefghi12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 0, 4, S("abcdefghi1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 0, 5, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 0, 6, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 1, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 1, 1, S("abcdefghi2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 1, 2, S("abcdefghi23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 1, 3, S("abcdefghi234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 1, 4, S("abcdefghi2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 1, 5, S("abcdefghi2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 2, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 2, 1, S("abcdefghi3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 2, 2, S("abcdefghi34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 2, 3, S("abcdefghi345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 2, 4, S("abcdefghi345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 4, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 4, 1, S("abcdefghi5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 4, 2, S("abcdefghi5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 5, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 5, 1, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 0, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 0, 1, S("abcdefghi1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 0, 5, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 0, 9, S("abcdefghi123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 0, 10, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 0, 11, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 1, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 1, 1, S("abcdefghi2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 1, 4, S("abcdefghi2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 1, 8, S("abcdefghi23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 1, 9, S("abcdefghi234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 1, 10, S("abcdefghi234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 5, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 5, 1, S("abcdefghi6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 5, 2, S("abcdefghi67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 5, 4, S("abcdefghi6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 5, 5, S("abcdefghi67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 5, 6, S("abcdefghi67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 9, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 9, 1, S("abcdefghi0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 9, 2, S("abcdefghi0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 10, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 10, 1, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 0, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 0, 1, S("abcdefghi1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 0, 10, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 0, 19, S("abcdefghi1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 0, 20, S("abcdefghi12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 0, 21, S("abcdefghi12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 1, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 1, 1, S("abcdefghi2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 1, 9, S("abcdefghi234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 1, 18, S("abcdefghi234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 1, 19, S("abcdefghi2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 1, 20, S("abcdefghi2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 10, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 10, 1, S("abcdefghi1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 10, 5, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 10, 9, S("abcdefghi123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 10, 10, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 10, 11, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 19, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 19, 1, S("abcdefghi0"))));
    return 0;
}

template <class S>
static int test34()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 19, 2, S("abcdefghi0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 20, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 20, 1, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S(""), 0, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S(""), 0, 1, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 0, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 0, 1, S("abcdefghi1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 0, 2, S("abcdefghi12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 0, 4, S("abcdefghi1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 0, 5, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 0, 6, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 1, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 1, 1, S("abcdefghi2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 1, 2, S("abcdefghi23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 1, 3, S("abcdefghi234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 1, 4, S("abcdefghi2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 1, 5, S("abcdefghi2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 2, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 2, 1, S("abcdefghi3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 2, 2, S("abcdefghi34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 2, 3, S("abcdefghi345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 2, 4, S("abcdefghi345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 4, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 4, 1, S("abcdefghi5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 4, 2, S("abcdefghi5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 5, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 5, 1, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 0, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 0, 1, S("abcdefghi1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 0, 5, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 0, 9, S("abcdefghi123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 0, 10, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 0, 11, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 1, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 1, 1, S("abcdefghi2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 1, 4, S("abcdefghi2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 1, 8, S("abcdefghi23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 1, 9, S("abcdefghi234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 1, 10, S("abcdefghi234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 5, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 5, 1, S("abcdefghi6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 5, 2, S("abcdefghi67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 5, 4, S("abcdefghi6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 5, 5, S("abcdefghi67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 5, 6, S("abcdefghi67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 9, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 9, 1, S("abcdefghi0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 9, 2, S("abcdefghi0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 10, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 10, 1, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 0, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 0, 1, S("abcdefghi1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 0, 10, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 0, 19, S("abcdefghi1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 0, 20, S("abcdefghi12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 0, 21, S("abcdefghi12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 1, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 1, 1, S("abcdefghi2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 1, 9, S("abcdefghi234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 1, 18, S("abcdefghi234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 1, 19, S("abcdefghi2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 1, 20, S("abcdefghi2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 10, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 10, 1, S("abcdefghi1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 10, 5, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 10, 9, S("abcdefghi123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 10, 10, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 10, 11, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 19, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 19, 1, S("abcdefghi0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 19, 2, S("abcdefghi0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 20, 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 20, 1, S("abcdefghi"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 9, 2, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S(""), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S(""), 0, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 0, 2, S("abcdefghij12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 0, 4, S("abcdefghij1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 0, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 0, 6, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 1, 2, S("abcdefghij23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 1, 3, S("abcdefghij234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 1, 4, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 1, 5, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 2, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 2, 1, S("abcdefghij3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 2, 2, S("abcdefghij34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 2, 3, S("abcdefghij345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 2, 4, S("abcdefghij345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 4, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 4, 1, S("abcdefghij5"))));
    return 0;
}

template <class S>
static int test35()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 4, 2, S("abcdefghij5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 5, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 0, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 0, 9, S("abcdefghij123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 0, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 0, 11, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 1, 4, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 1, 8, S("abcdefghij23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 1, 9, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 1, 10, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 5, 1, S("abcdefghij6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 5, 2, S("abcdefghij67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 5, 4, S("abcdefghij6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 5, 5, S("abcdefghij67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 5, 6, S("abcdefghij67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 9, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 9, 1, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 9, 2, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 10, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 0, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 0, 19, S("abcdefghij1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 0, 20, S("abcdefghij12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 0, 21, S("abcdefghij12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 1, 9, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 1, 18, S("abcdefghij234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 1, 19, S("abcdefghij2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 1, 20, S("abcdefghij2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 10, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 10, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 10, 9, S("abcdefghij123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 10, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 10, 11, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 19, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 19, 1, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 19, 2, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 20, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 20, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S(""), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S(""), 0, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 0, 2, S("abcdefghij12"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 0, 4, S("abcdefghij1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 0, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 0, 6, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 1, 2, S("abcdefghij23"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 1, 3, S("abcdefghij234"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 1, 4, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 1, 5, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 2, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 2, 1, S("abcdefghij3"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 2, 2, S("abcdefghij34"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 2, 3, S("abcdefghij345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 2, 4, S("abcdefghij345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 4, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 4, 1, S("abcdefghij5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 4, 2, S("abcdefghij5"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 5, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 0, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 0, 9, S("abcdefghij123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 0, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 0, 11, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 1, 4, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 1, 8, S("abcdefghij23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 1, 9, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 1, 10, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 5, 1, S("abcdefghij6"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 5, 2, S("abcdefghij67"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 5, 4, S("abcdefghij6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 5, 5, S("abcdefghij67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 5, 6, S("abcdefghij67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 9, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 9, 1, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 9, 2, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 10, 0, S("abcdefghij"))));
    return 0;
}

template <class S>
static int test36()
{
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 10, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 0, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 0, 19, S("abcdefghij1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 0, 20, S("abcdefghij12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 0, 21, S("abcdefghij12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 1, 9, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 1, 18, S("abcdefghij234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 1, 19, S("abcdefghij2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 1, 20, S("abcdefghij2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 10, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 10, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 10, 9, S("abcdefghij123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 10, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 10, 11, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 19, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 19, 1, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 19, 2, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 20, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 20, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 10, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S(""), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S(""), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 0, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 0, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 0, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 0, 6, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 1, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 1, 3, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 1, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 1, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 2, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 2, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 2, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 2, 3, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 2, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 4, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 4, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 4, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 5, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 5, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 0, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 0, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 0, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 0, 11, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 1, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 1, 8, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 1, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 1, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 5, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 5, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 5, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 5, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 5, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 5, 6, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 9, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 9, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 9, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 10, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 10, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 0, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 0, 19, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 0, 20, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 0, 21, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 1, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 1, 18, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 1, 19, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 1, 20, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 10, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 10, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 10, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 10, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 10, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 10, 11, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 19, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 19, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 19, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 20, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 20, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghij"), 11, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    return 0;
}

template <class S>
static int test37()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S(""), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S(""), 0, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 0, 1, S("1abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 0, 2, S("12abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 0, 4, S("1234abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 0, 5, S("12345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 0, 6, S("12345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 1, 1, S("2abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 1, 2, S("23abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 1, 3, S("234abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 1, 4, S("2345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 1, 5, S("2345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 2, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 2, 1, S("3abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 2, 2, S("34abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 2, 3, S("345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 2, 4, S("345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 4, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 4, 1, S("5abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 4, 2, S("5abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 5, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 0, 1, S("1abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 0, 5, S("12345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 0, 9, S("123456789abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 0, 10, S("1234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 0, 11, S("1234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 1, 1, S("2abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 1, 4, S("2345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 1, 8, S("23456789abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 1, 9, S("234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 1, 10, S("234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 5, 1, S("6abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 5, 2, S("67abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 5, 4, S("6789abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 5, 5, S("67890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 5, 6, S("67890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 9, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 9, 1, S("0abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 9, 2, S("0abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 10, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 0, 1, S("1abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 0, 10, S("1234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 0, 19, S("1234567890123456789abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 0, 20, S("12345678901234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 0, 21, S("12345678901234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 1, 1, S("2abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 1, 9, S("234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 1, 18, S("234567890123456789abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 1, 19, S("2345678901234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 1, 20, S("2345678901234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 10, 1, S("1abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 10, 5, S("12345abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 10, 9, S("123456789abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 10, 10, S("1234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 10, 11, S("1234567890abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 19, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 19, 1, S("0abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 19, 2, S("0abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 20, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 20, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S(""), 0, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S(""), 0, 1, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 0, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 0, 1, S("1bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 0, 2, S("12bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 0, 4, S("1234bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 0, 5, S("12345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 0, 6, S("12345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 1, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 1, 1, S("2bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 1, 2, S("23bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 1, 3, S("234bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 1, 4, S("2345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 1, 5, S("2345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 2, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 2, 1, S("3bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 2, 2, S("34bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 2, 3, S("345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 2, 4, S("345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 4, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 4, 1, S("5bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 4, 2, S("5bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 5, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 5, 1, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345"), 6, 0, S("can't happen"))));
    return 0;
}

template <class S>
static int test38()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 0, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 0, 1, S("1bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 0, 5, S("12345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 0, 9, S("123456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 0, 10, S("1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 0, 11, S("1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 1, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 1, 1, S("2bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 1, 4, S("2345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 1, 8, S("23456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 1, 9, S("234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 1, 10, S("234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 5, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 5, 1, S("6bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 5, 2, S("67bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 5, 4, S("6789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 5, 5, S("67890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 5, 6, S("67890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 9, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 9, 1, S("0bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 9, 2, S("0bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 10, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 10, 1, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 0, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 0, 1, S("1bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 0, 10, S("1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 0, 19, S("1234567890123456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 0, 20, S("12345678901234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 0, 21, S("12345678901234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 1, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 1, 1, S("2bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 1, 9, S("234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 1, 18, S("234567890123456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 1, 19, S("2345678901234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 1, 20, S("2345678901234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 10, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 10, 1, S("1bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 10, 5, S("12345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 10, 9, S("123456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 10, 10, S("1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 10, 11, S("1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 19, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 19, 1, S("0bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 19, 2, S("0bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 20, 0, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 20, 1, S("bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S(""), 0, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S(""), 0, 1, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 0, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 0, 1, S("1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 0, 2, S("12klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 0, 4, S("1234klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 0, 5, S("12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 0, 6, S("12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 1, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 1, 1, S("2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 1, 2, S("23klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 1, 3, S("234klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 1, 4, S("2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 1, 5, S("2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 2, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 2, 1, S("3klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 2, 2, S("34klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 2, 3, S("345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 2, 4, S("345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 4, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 4, 1, S("5klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 4, 2, S("5klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 5, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 5, 1, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 0, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 0, 1, S("1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 0, 5, S("12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 0, 9, S("123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 0, 10, S("1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 0, 11, S("1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 1, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 1, 1, S("2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 1, 4, S("2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 1, 8, S("23456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 1, 9, S("234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 1, 10, S("234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 5, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 5, 1, S("6klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 5, 2, S("67klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 5, 4, S("6789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 5, 5, S("67890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 5, 6, S("67890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 9, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 9, 1, S("0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 9, 2, S("0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 10, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 10, 1, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 0, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 0, 1, S("1klmnopqrst"))));
    return 0;
}

template <class S>
static int test39()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 0, 10, S("1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 0, 19, S("1234567890123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 0, 20, S("12345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 0, 21, S("12345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 1, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 1, 1, S("2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 1, 9, S("234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 1, 18, S("234567890123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 1, 19, S("2345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 1, 20, S("2345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 10, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 10, 1, S("1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 10, 5, S("12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 10, 9, S("123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 10, 10, S("1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 10, 11, S("1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 19, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 19, 1, S("0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 19, 2, S("0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 20, 0, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 20, 1, S("klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 10, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S(""), 0, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S(""), 0, 1, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 0, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 0, 1, S("1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 0, 2, S("12t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 0, 4, S("1234t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 0, 5, S("12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 0, 6, S("12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 1, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 1, 1, S("2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 1, 2, S("23t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 1, 3, S("234t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 1, 4, S("2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 1, 5, S("2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 2, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 2, 1, S("3t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 2, 2, S("34t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 2, 3, S("345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 2, 4, S("345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 4, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 4, 1, S("5t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 4, 2, S("5t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 5, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 5, 1, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 0, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 0, 1, S("1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 0, 5, S("12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 0, 9, S("123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 0, 10, S("1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 0, 11, S("1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 1, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 1, 1, S("2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 1, 4, S("2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 1, 8, S("23456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 1, 9, S("234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 1, 10, S("234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 5, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 5, 1, S("6t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 5, 2, S("67t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 5, 4, S("6789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 5, 5, S("67890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 5, 6, S("67890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 9, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 9, 1, S("0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 9, 2, S("0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 10, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 10, 1, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 0, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 0, 1, S("1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 0, 10, S("1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 0, 19, S("1234567890123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 0, 20, S("12345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 0, 21, S("12345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 1, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 1, 1, S("2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 1, 9, S("234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 1, 18, S("234567890123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 1, 19, S("2345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 1, 20, S("2345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 10, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 10, 1, S("1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 10, 5, S("12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 10, 9, S("123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 10, 10, S("1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 10, 11, S("1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 19, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 19, 1, S("0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 19, 2, S("0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 20, 0, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 20, 1, S("t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 19, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S(""), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S(""), 0, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 0, 0, S(""))));
    return 0;
}

template <class S>
static int test40()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 0, 2, S("12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 0, 4, S("1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 0, 6, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 1, 2, S("23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 1, 3, S("234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 1, 5, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 2, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 2, 1, S("3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 2, 2, S("34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 2, 3, S("345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 2, 4, S("345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 4, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 4, 1, S("5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 4, 2, S("5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 5, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 0, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 0, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 1, 8, S("23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 1, 10, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 5, 1, S("6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 5, 2, S("67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 5, 4, S("6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 5, 5, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 5, 6, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 9, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 9, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 9, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 10, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 0, 19, S("1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 0, 20, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 0, 21, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 1, 18, S("234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 1, 19, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 1, 20, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 10, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 10, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 10, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 10, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 10, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 19, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 19, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 19, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 20, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 20, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 20, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S(""), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S(""), 0, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 0, 2, S("12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 0, 4, S("1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 0, 6, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 1, 2, S("23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 1, 3, S("234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 1, 5, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 2, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 2, 1, S("3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 2, 2, S("34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 2, 3, S("345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 2, 4, S("345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 4, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 4, 1, S("5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 4, 2, S("5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 5, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 0, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 0, 9, S("123456789"))));
    return 0;
}

template <class S>
static int test41()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 0, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 1, 4, S("2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 1, 8, S("23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 1, 10, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 5, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 5, 1, S("6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 5, 2, S("67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 5, 4, S("6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 5, 5, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 5, 6, S("67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 9, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 9, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 9, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 10, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 0, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 0, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 0, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 0, 19, S("1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 0, 20, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 0, 21, S("12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 1, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 1, 1, S("2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 1, 9, S("234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 1, 18, S("234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 1, 19, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 1, 20, S("2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 10, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 10, 1, S("1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 10, 5, S("12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 10, 9, S("123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 10, 10, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 10, 11, S("1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 19, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 19, 1, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 19, 2, S("0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 20, 0, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 20, 1, S(""))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 0, 21, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S(""), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S(""), 0, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 0, 1, S("a1bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 0, 2, S("a12bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 0, 4, S("a1234bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 0, 5, S("a12345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 0, 6, S("a12345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 1, 1, S("a2bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 1, 2, S("a23bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 1, 3, S("a234bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 1, 4, S("a2345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 1, 5, S("a2345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 2, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 2, 1, S("a3bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 2, 2, S("a34bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 2, 3, S("a345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 2, 4, S("a345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 4, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 4, 1, S("a5bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 4, 2, S("a5bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 5, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 0, 1, S("a1bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 0, 5, S("a12345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 0, 9, S("a123456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 0, 10, S("a1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 0, 11, S("a1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 1, 1, S("a2bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 1, 4, S("a2345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 1, 8, S("a23456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 1, 9, S("a234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 1, 10, S("a234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 5, 1, S("a6bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 5, 2, S("a67bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 5, 4, S("a6789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 5, 5, S("a67890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 5, 6, S("a67890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 9, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 9, 1, S("a0bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 9, 2, S("a0bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 10, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 0, 1, S("a1bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 0, 10, S("a1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 0, 19, S("a1234567890123456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 0, 20, S("a12345678901234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 0, 21, S("a12345678901234567890bcdefghijklmnopqrst"))));
    return 0;
}

template <class S>
static int test42()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 1, 1, S("a2bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 1, 9, S("a234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 1, 18, S("a234567890123456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 1, 19, S("a2345678901234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 1, 20, S("a2345678901234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 10, 1, S("a1bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 10, 5, S("a12345bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 10, 9, S("a123456789bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 10, 10, S("a1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 10, 11, S("a1234567890bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 19, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 19, 1, S("a0bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 19, 2, S("a0bcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 20, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 20, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S(""), 0, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S(""), 0, 1, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 0, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 0, 1, S("a1cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 0, 2, S("a12cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 0, 4, S("a1234cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 0, 5, S("a12345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 0, 6, S("a12345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 1, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 1, 1, S("a2cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 1, 2, S("a23cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 1, 3, S("a234cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 1, 4, S("a2345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 1, 5, S("a2345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 2, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 2, 1, S("a3cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 2, 2, S("a34cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 2, 3, S("a345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 2, 4, S("a345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 4, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 4, 1, S("a5cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 4, 2, S("a5cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 5, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 5, 1, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 0, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 0, 1, S("a1cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 0, 5, S("a12345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 0, 9, S("a123456789cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 0, 10, S("a1234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 0, 11, S("a1234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 1, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 1, 1, S("a2cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 1, 4, S("a2345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 1, 8, S("a23456789cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 1, 9, S("a234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 1, 10, S("a234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 5, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 5, 1, S("a6cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 5, 2, S("a67cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 5, 4, S("a6789cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 5, 5, S("a67890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 5, 6, S("a67890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 9, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 9, 1, S("a0cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 9, 2, S("a0cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 10, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 10, 1, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 0, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 0, 1, S("a1cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 0, 10, S("a1234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 0, 19, S("a1234567890123456789cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 0, 20, S("a12345678901234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 0, 21, S("a12345678901234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 1, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 1, 1, S("a2cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 1, 9, S("a234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 1, 18, S("a234567890123456789cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 1, 19, S("a2345678901234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 1, 20, S("a2345678901234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 10, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 10, 1, S("a1cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 10, 5, S("a12345cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 10, 9, S("a123456789cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 10, 10, S("a1234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 10, 11, S("a1234567890cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 19, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 19, 1, S("a0cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 19, 2, S("a0cdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 20, 0, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 20, 1, S("acdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S(""), 0, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S(""), 0, 1, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 0, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 0, 1, S("a1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 0, 2, S("a12klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 0, 4, S("a1234klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 0, 5, S("a12345klmnopqrst"))));
    return 0;
}

template <class S>
static int test43()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 0, 6, S("a12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 1, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 1, 1, S("a2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 1, 2, S("a23klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 1, 3, S("a234klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 1, 4, S("a2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 1, 5, S("a2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 2, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 2, 1, S("a3klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 2, 2, S("a34klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 2, 3, S("a345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 2, 4, S("a345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 4, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 4, 1, S("a5klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 4, 2, S("a5klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 5, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 5, 1, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 0, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 0, 1, S("a1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 0, 5, S("a12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 0, 9, S("a123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 0, 10, S("a1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 0, 11, S("a1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 1, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 1, 1, S("a2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 1, 4, S("a2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 1, 8, S("a23456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 1, 9, S("a234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 1, 10, S("a234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 5, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 5, 1, S("a6klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 5, 2, S("a67klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 5, 4, S("a6789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 5, 5, S("a67890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 5, 6, S("a67890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 9, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 9, 1, S("a0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 9, 2, S("a0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 10, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 10, 1, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 0, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 0, 1, S("a1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 0, 10, S("a1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 0, 19, S("a1234567890123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 0, 20, S("a12345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 0, 21, S("a12345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 1, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 1, 1, S("a2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 1, 9, S("a234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 1, 18, S("a234567890123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 1, 19, S("a2345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 1, 20, S("a2345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 10, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 10, 1, S("a1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 10, 5, S("a12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 10, 9, S("a123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 10, 10, S("a1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 10, 11, S("a1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 19, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 19, 1, S("a0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 19, 2, S("a0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 20, 0, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 20, 1, S("aklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 9, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S(""), 0, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S(""), 0, 1, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 0, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 0, 1, S("a1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 0, 2, S("a12t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 0, 4, S("a1234t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 0, 5, S("a12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 0, 6, S("a12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 1, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 1, 1, S("a2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 1, 2, S("a23t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 1, 3, S("a234t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 1, 4, S("a2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 1, 5, S("a2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 2, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 2, 1, S("a3t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 2, 2, S("a34t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 2, 3, S("a345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 2, 4, S("a345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 4, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 4, 1, S("a5t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 4, 2, S("a5t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 5, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 5, 1, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 0, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 0, 1, S("a1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 0, 5, S("a12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 0, 9, S("a123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 0, 10, S("a1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 0, 11, S("a1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 1, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 1, 1, S("a2t"))));
    return 0;
}

template <class S>
static int test44()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 1, 4, S("a2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 1, 8, S("a23456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 1, 9, S("a234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 1, 10, S("a234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 5, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 5, 1, S("a6t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 5, 2, S("a67t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 5, 4, S("a6789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 5, 5, S("a67890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 5, 6, S("a67890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 9, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 9, 1, S("a0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 9, 2, S("a0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 10, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 10, 1, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 0, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 0, 1, S("a1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 0, 10, S("a1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 0, 19, S("a1234567890123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 0, 20, S("a12345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 0, 21, S("a12345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 1, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 1, 1, S("a2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 1, 9, S("a234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 1, 18, S("a234567890123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 1, 19, S("a2345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 1, 20, S("a2345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 10, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 10, 1, S("a1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 10, 5, S("a12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 10, 9, S("a123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 10, 10, S("a1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 10, 11, S("a1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 19, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 19, 1, S("a0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 19, 2, S("a0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 20, 0, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 20, 1, S("at"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 18, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S(""), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S(""), 0, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 0, 2, S("a12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 0, 4, S("a1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 0, 6, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 1, 2, S("a23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 1, 3, S("a234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 1, 5, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 2, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 2, 1, S("a3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 2, 2, S("a34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 2, 3, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 2, 4, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 4, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 4, 1, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 4, 2, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 5, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 0, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 0, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 1, 8, S("a23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 1, 10, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 5, 1, S("a6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 5, 2, S("a67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 5, 4, S("a6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 5, 5, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 5, 6, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 9, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 9, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 9, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 10, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 0, 19, S("a1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 0, 20, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 0, 21, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 1, 18, S("a234567890123456789"))));
    return 0;
}

template <class S>
static int test45()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 1, 19, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 1, 20, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 10, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 10, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 10, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 10, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 10, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 19, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 19, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 19, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 20, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 20, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 19, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S(""), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S(""), 0, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 0, 2, S("a12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 0, 4, S("a1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 0, 6, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 1, 2, S("a23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 1, 3, S("a234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 1, 5, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 2, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 2, 1, S("a3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 2, 2, S("a34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 2, 3, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 2, 4, S("a345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 4, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 4, 1, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 4, 2, S("a5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 5, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 0, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 0, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 0, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 1, 4, S("a2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 1, 8, S("a23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 1, 10, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 5, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 5, 1, S("a6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 5, 2, S("a67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 5, 4, S("a6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 5, 5, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 5, 6, S("a67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 9, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 9, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 9, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 10, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 0, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 0, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 0, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 0, 19, S("a1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 0, 20, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 0, 21, S("a12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 1, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 1, 1, S("a2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 1, 9, S("a234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 1, 18, S("a234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 1, 19, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 1, 20, S("a2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 10, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 10, 1, S("a1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 10, 5, S("a12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 10, 9, S("a123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 10, 10, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 10, 11, S("a1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 19, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 19, 1, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 19, 2, S("a0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 20, 0, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 20, 1, S("a"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 1, 20, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S(""), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S(""), 0, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 0, 1, S("abcdefghij1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 0, 2, S("abcdefghij12klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 0, 4, S("abcdefghij1234klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 0, 5, S("abcdefghij12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 0, 6, S("abcdefghij12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 1, 1, S("abcdefghij2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 1, 2, S("abcdefghij23klmnopqrst"))));
    return 0;
}

template <class S>
static int test46()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 1, 3, S("abcdefghij234klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 1, 4, S("abcdefghij2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 1, 5, S("abcdefghij2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 2, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 2, 1, S("abcdefghij3klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 2, 2, S("abcdefghij34klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 2, 3, S("abcdefghij345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 2, 4, S("abcdefghij345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 4, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 4, 1, S("abcdefghij5klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 4, 2, S("abcdefghij5klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 5, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 0, 1, S("abcdefghij1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 0, 5, S("abcdefghij12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 0, 9, S("abcdefghij123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 0, 10, S("abcdefghij1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 0, 11, S("abcdefghij1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 1, 1, S("abcdefghij2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 1, 4, S("abcdefghij2345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 1, 8, S("abcdefghij23456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 1, 9, S("abcdefghij234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 1, 10, S("abcdefghij234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 5, 1, S("abcdefghij6klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 5, 2, S("abcdefghij67klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 5, 4, S("abcdefghij6789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 5, 5, S("abcdefghij67890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 5, 6, S("abcdefghij67890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 9, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 9, 1, S("abcdefghij0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 9, 2, S("abcdefghij0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 10, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 0, 1, S("abcdefghij1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 0, 10, S("abcdefghij1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 0, 19, S("abcdefghij1234567890123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 0, 20, S("abcdefghij12345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 0, 21, S("abcdefghij12345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 1, 1, S("abcdefghij2klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 1, 9, S("abcdefghij234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 1, 18, S("abcdefghij234567890123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 1, 19, S("abcdefghij2345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 1, 20, S("abcdefghij2345678901234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 10, 1, S("abcdefghij1klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 10, 5, S("abcdefghij12345klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 10, 9, S("abcdefghij123456789klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 10, 10, S("abcdefghij1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 10, 11, S("abcdefghij1234567890klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 19, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 19, 1, S("abcdefghij0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 19, 2, S("abcdefghij0klmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 20, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 20, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S(""), 0, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S(""), 0, 1, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 0, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 0, 1, S("abcdefghij1lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 0, 2, S("abcdefghij12lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 0, 4, S("abcdefghij1234lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 0, 5, S("abcdefghij12345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 0, 6, S("abcdefghij12345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 1, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 1, 1, S("abcdefghij2lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 1, 2, S("abcdefghij23lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 1, 3, S("abcdefghij234lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 1, 4, S("abcdefghij2345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 1, 5, S("abcdefghij2345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 2, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 2, 1, S("abcdefghij3lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 2, 2, S("abcdefghij34lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 2, 3, S("abcdefghij345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 2, 4, S("abcdefghij345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 4, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 4, 1, S("abcdefghij5lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 4, 2, S("abcdefghij5lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 5, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 5, 1, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 0, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 0, 1, S("abcdefghij1lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 0, 5, S("abcdefghij12345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 0, 9, S("abcdefghij123456789lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 0, 10, S("abcdefghij1234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 0, 11, S("abcdefghij1234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 1, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 1, 1, S("abcdefghij2lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 1, 4, S("abcdefghij2345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 1, 8, S("abcdefghij23456789lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 1, 9, S("abcdefghij234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 1, 10, S("abcdefghij234567890lmnopqrst"))));
    return 0;
}

template <class S>
static int test47()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 5, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 5, 1, S("abcdefghij6lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 5, 2, S("abcdefghij67lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 5, 4, S("abcdefghij6789lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 5, 5, S("abcdefghij67890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 5, 6, S("abcdefghij67890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 9, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 9, 1, S("abcdefghij0lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 9, 2, S("abcdefghij0lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 10, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 10, 1, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 0, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 0, 1, S("abcdefghij1lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 0, 10, S("abcdefghij1234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 0, 19, S("abcdefghij1234567890123456789lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 0, 20, S("abcdefghij12345678901234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 0, 21, S("abcdefghij12345678901234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 1, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 1, 1, S("abcdefghij2lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 1, 9, S("abcdefghij234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 1, 18, S("abcdefghij234567890123456789lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 1, 19, S("abcdefghij2345678901234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 1, 20, S("abcdefghij2345678901234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 10, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 10, 1, S("abcdefghij1lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 10, 5, S("abcdefghij12345lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 10, 9, S("abcdefghij123456789lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 10, 10, S("abcdefghij1234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 10, 11, S("abcdefghij1234567890lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 19, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 19, 1, S("abcdefghij0lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 19, 2, S("abcdefghij0lmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 20, 0, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 20, 1, S("abcdefghijlmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S(""), 0, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S(""), 0, 1, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 0, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 0, 1, S("abcdefghij1pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 0, 2, S("abcdefghij12pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 0, 4, S("abcdefghij1234pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 0, 5, S("abcdefghij12345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 0, 6, S("abcdefghij12345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 1, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 1, 1, S("abcdefghij2pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 1, 2, S("abcdefghij23pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 1, 3, S("abcdefghij234pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 1, 4, S("abcdefghij2345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 1, 5, S("abcdefghij2345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 2, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 2, 1, S("abcdefghij3pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 2, 2, S("abcdefghij34pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 2, 3, S("abcdefghij345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 2, 4, S("abcdefghij345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 4, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 4, 1, S("abcdefghij5pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 4, 2, S("abcdefghij5pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 5, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 5, 1, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 0, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 0, 1, S("abcdefghij1pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 0, 5, S("abcdefghij12345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 0, 9, S("abcdefghij123456789pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 0, 10, S("abcdefghij1234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 0, 11, S("abcdefghij1234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 1, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 1, 1, S("abcdefghij2pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 1, 4, S("abcdefghij2345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 1, 8, S("abcdefghij23456789pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 1, 9, S("abcdefghij234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 1, 10, S("abcdefghij234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 5, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 5, 1, S("abcdefghij6pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 5, 2, S("abcdefghij67pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 5, 4, S("abcdefghij6789pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 5, 5, S("abcdefghij67890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 5, 6, S("abcdefghij67890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 9, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 9, 1, S("abcdefghij0pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 9, 2, S("abcdefghij0pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 10, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 10, 1, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 0, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 0, 1, S("abcdefghij1pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 0, 10, S("abcdefghij1234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 0, 19, S("abcdefghij1234567890123456789pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 0, 20, S("abcdefghij12345678901234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 0, 21, S("abcdefghij12345678901234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 1, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 1, 1, S("abcdefghij2pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 1, 9, S("abcdefghij234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 1, 18, S("abcdefghij234567890123456789pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 1, 19, S("abcdefghij2345678901234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 1, 20, S("abcdefghij2345678901234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 10, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 10, 1, S("abcdefghij1pqrst"))));
    return 0;
}

template <class S>
static int test48()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 10, 5, S("abcdefghij12345pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 10, 9, S("abcdefghij123456789pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 10, 10, S("abcdefghij1234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 10, 11, S("abcdefghij1234567890pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 19, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 19, 1, S("abcdefghij0pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 19, 2, S("abcdefghij0pqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 20, 0, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 20, 1, S("abcdefghijpqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 5, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S(""), 0, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S(""), 0, 1, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 0, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 0, 1, S("abcdefghij1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 0, 2, S("abcdefghij12t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 0, 4, S("abcdefghij1234t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 0, 5, S("abcdefghij12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 0, 6, S("abcdefghij12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 1, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 1, 1, S("abcdefghij2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 1, 2, S("abcdefghij23t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 1, 3, S("abcdefghij234t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 1, 4, S("abcdefghij2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 1, 5, S("abcdefghij2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 2, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 2, 1, S("abcdefghij3t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 2, 2, S("abcdefghij34t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 2, 3, S("abcdefghij345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 2, 4, S("abcdefghij345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 4, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 4, 1, S("abcdefghij5t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 4, 2, S("abcdefghij5t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 5, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 5, 1, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 0, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 0, 1, S("abcdefghij1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 0, 5, S("abcdefghij12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 0, 9, S("abcdefghij123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 0, 10, S("abcdefghij1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 0, 11, S("abcdefghij1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 1, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 1, 1, S("abcdefghij2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 1, 4, S("abcdefghij2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 1, 8, S("abcdefghij23456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 1, 9, S("abcdefghij234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 1, 10, S("abcdefghij234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 5, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 5, 1, S("abcdefghij6t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 5, 2, S("abcdefghij67t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 5, 4, S("abcdefghij6789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 5, 5, S("abcdefghij67890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 5, 6, S("abcdefghij67890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 9, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 9, 1, S("abcdefghij0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 9, 2, S("abcdefghij0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 10, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 10, 1, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 0, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 0, 1, S("abcdefghij1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 0, 10, S("abcdefghij1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 0, 19, S("abcdefghij1234567890123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 0, 20, S("abcdefghij12345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 0, 21, S("abcdefghij12345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 1, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 1, 1, S("abcdefghij2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 1, 9, S("abcdefghij234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 1, 18, S("abcdefghij234567890123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 1, 19, S("abcdefghij2345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 1, 20, S("abcdefghij2345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 10, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 10, 1, S("abcdefghij1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 10, 5, S("abcdefghij12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 10, 9, S("abcdefghij123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 10, 10, S("abcdefghij1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 10, 11, S("abcdefghij1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 19, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 19, 1, S("abcdefghij0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 19, 2, S("abcdefghij0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 20, 0, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 20, 1, S("abcdefghijt"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 9, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S(""), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S(""), 0, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 0, 2, S("abcdefghij12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 0, 4, S("abcdefghij1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 0, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 0, 6, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 1, 2, S("abcdefghij23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 1, 3, S("abcdefghij234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 1, 4, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 1, 5, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 2, 0, S("abcdefghij"))));
    return 0;
}

template <class S>
static int test49()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 2, 1, S("abcdefghij3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 2, 2, S("abcdefghij34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 2, 3, S("abcdefghij345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 2, 4, S("abcdefghij345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 4, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 4, 1, S("abcdefghij5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 4, 2, S("abcdefghij5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 5, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 0, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 0, 9, S("abcdefghij123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 0, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 0, 11, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 1, 4, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 1, 8, S("abcdefghij23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 1, 9, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 1, 10, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 5, 1, S("abcdefghij6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 5, 2, S("abcdefghij67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 5, 4, S("abcdefghij6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 5, 5, S("abcdefghij67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 5, 6, S("abcdefghij67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 9, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 9, 1, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 9, 2, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 10, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 0, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 0, 19, S("abcdefghij1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 0, 20, S("abcdefghij12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 0, 21, S("abcdefghij12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 1, 9, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 1, 18, S("abcdefghij234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 1, 19, S("abcdefghij2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 1, 20, S("abcdefghij2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 10, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 10, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 10, 9, S("abcdefghij123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 10, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 10, 11, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 19, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 19, 1, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 19, 2, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 20, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 20, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 10, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S(""), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S(""), 0, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 0, 2, S("abcdefghij12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 0, 4, S("abcdefghij1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 0, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 0, 6, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 1, 2, S("abcdefghij23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 1, 3, S("abcdefghij234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 1, 4, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 1, 5, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 2, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 2, 1, S("abcdefghij3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 2, 2, S("abcdefghij34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 2, 3, S("abcdefghij345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 2, 4, S("abcdefghij345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 4, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 4, 1, S("abcdefghij5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 4, 2, S("abcdefghij5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 5, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 0, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 0, 9, S("abcdefghij123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 0, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 0, 11, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 1, 4, S("abcdefghij2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 1, 8, S("abcdefghij23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 1, 9, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 1, 10, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 5, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 5, 1, S("abcdefghij6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 5, 2, S("abcdefghij67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 5, 4, S("abcdefghij6789"))));
    return 0;
}

template <class S>
static int test50()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 5, 5, S("abcdefghij67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 5, 6, S("abcdefghij67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 9, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 9, 1, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 9, 2, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 10, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 0, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 0, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 0, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 0, 19, S("abcdefghij1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 0, 20, S("abcdefghij12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 0, 21, S("abcdefghij12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 1, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 1, 1, S("abcdefghij2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 1, 9, S("abcdefghij234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 1, 18, S("abcdefghij234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 1, 19, S("abcdefghij2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 1, 20, S("abcdefghij2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 10, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 10, 1, S("abcdefghij1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 10, 5, S("abcdefghij12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 10, 9, S("abcdefghij123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 10, 10, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 10, 11, S("abcdefghij1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 19, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 19, 1, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 19, 2, S("abcdefghij0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 20, 0, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 20, 1, S("abcdefghij"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 10, 11, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S(""), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S(""), 0, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 0, 1, S("abcdefghijklmnopqrs1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 0, 2, S("abcdefghijklmnopqrs12t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 0, 4, S("abcdefghijklmnopqrs1234t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 0, 5, S("abcdefghijklmnopqrs12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 0, 6, S("abcdefghijklmnopqrs12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 1, 1, S("abcdefghijklmnopqrs2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 1, 2, S("abcdefghijklmnopqrs23t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 1, 3, S("abcdefghijklmnopqrs234t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 1, 4, S("abcdefghijklmnopqrs2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 1, 5, S("abcdefghijklmnopqrs2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 2, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 2, 1, S("abcdefghijklmnopqrs3t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 2, 2, S("abcdefghijklmnopqrs34t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 2, 3, S("abcdefghijklmnopqrs345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 2, 4, S("abcdefghijklmnopqrs345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 4, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 4, 1, S("abcdefghijklmnopqrs5t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 4, 2, S("abcdefghijklmnopqrs5t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 5, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 0, 1, S("abcdefghijklmnopqrs1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 0, 5, S("abcdefghijklmnopqrs12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 0, 9, S("abcdefghijklmnopqrs123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 0, 10, S("abcdefghijklmnopqrs1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 0, 11, S("abcdefghijklmnopqrs1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 1, 1, S("abcdefghijklmnopqrs2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 1, 4, S("abcdefghijklmnopqrs2345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 1, 8, S("abcdefghijklmnopqrs23456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 1, 9, S("abcdefghijklmnopqrs234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 1, 10, S("abcdefghijklmnopqrs234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 5, 1, S("abcdefghijklmnopqrs6t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 5, 2, S("abcdefghijklmnopqrs67t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 5, 4, S("abcdefghijklmnopqrs6789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 5, 5, S("abcdefghijklmnopqrs67890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 5, 6, S("abcdefghijklmnopqrs67890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 9, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 9, 1, S("abcdefghijklmnopqrs0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 9, 2, S("abcdefghijklmnopqrs0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 10, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 0, 1, S("abcdefghijklmnopqrs1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 0, 10, S("abcdefghijklmnopqrs1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 0, 19, S("abcdefghijklmnopqrs1234567890123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 0, 20, S("abcdefghijklmnopqrs12345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 0, 21, S("abcdefghijklmnopqrs12345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 1, 1, S("abcdefghijklmnopqrs2t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 1, 9, S("abcdefghijklmnopqrs234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 1, 18, S("abcdefghijklmnopqrs234567890123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 1, 19, S("abcdefghijklmnopqrs2345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 1, 20, S("abcdefghijklmnopqrs2345678901234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 10, 1, S("abcdefghijklmnopqrs1t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 10, 5, S("abcdefghijklmnopqrs12345t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 10, 9, S("abcdefghijklmnopqrs123456789t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 10, 10, S("abcdefghijklmnopqrs1234567890t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 10, 11, S("abcdefghijklmnopqrs1234567890t"))));
    return 0;
}

template <class S>
static int test51()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 19, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 19, 1, S("abcdefghijklmnopqrs0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 19, 2, S("abcdefghijklmnopqrs0t"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 20, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 20, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S(""), 0, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S(""), 0, 1, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 0, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 0, 1, S("abcdefghijklmnopqrs1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 0, 2, S("abcdefghijklmnopqrs12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 0, 4, S("abcdefghijklmnopqrs1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 0, 5, S("abcdefghijklmnopqrs12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 0, 6, S("abcdefghijklmnopqrs12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 1, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 1, 1, S("abcdefghijklmnopqrs2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 1, 2, S("abcdefghijklmnopqrs23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 1, 3, S("abcdefghijklmnopqrs234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 1, 4, S("abcdefghijklmnopqrs2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 1, 5, S("abcdefghijklmnopqrs2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 2, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 2, 1, S("abcdefghijklmnopqrs3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 2, 2, S("abcdefghijklmnopqrs34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 2, 3, S("abcdefghijklmnopqrs345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 2, 4, S("abcdefghijklmnopqrs345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 4, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 4, 1, S("abcdefghijklmnopqrs5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 4, 2, S("abcdefghijklmnopqrs5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 5, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 5, 1, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 0, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 0, 1, S("abcdefghijklmnopqrs1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 0, 5, S("abcdefghijklmnopqrs12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 0, 9, S("abcdefghijklmnopqrs123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 0, 10, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 0, 11, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 1, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 1, 1, S("abcdefghijklmnopqrs2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 1, 4, S("abcdefghijklmnopqrs2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 1, 8, S("abcdefghijklmnopqrs23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 1, 9, S("abcdefghijklmnopqrs234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 1, 10, S("abcdefghijklmnopqrs234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 5, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 5, 1, S("abcdefghijklmnopqrs6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 5, 2, S("abcdefghijklmnopqrs67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 5, 4, S("abcdefghijklmnopqrs6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 5, 5, S("abcdefghijklmnopqrs67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 5, 6, S("abcdefghijklmnopqrs67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 9, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 9, 1, S("abcdefghijklmnopqrs0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 9, 2, S("abcdefghijklmnopqrs0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 10, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 10, 1, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 0, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 0, 1, S("abcdefghijklmnopqrs1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 0, 10, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 0, 19, S("abcdefghijklmnopqrs1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 0, 20, S("abcdefghijklmnopqrs12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 0, 21, S("abcdefghijklmnopqrs12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 1, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 1, 1, S("abcdefghijklmnopqrs2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 1, 9, S("abcdefghijklmnopqrs234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 1, 18, S("abcdefghijklmnopqrs234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 1, 19, S("abcdefghijklmnopqrs2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 1, 20, S("abcdefghijklmnopqrs2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 10, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 10, 1, S("abcdefghijklmnopqrs1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 10, 5, S("abcdefghijklmnopqrs12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 10, 9, S("abcdefghijklmnopqrs123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 10, 10, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 10, 11, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 19, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 19, 1, S("abcdefghijklmnopqrs0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 19, 2, S("abcdefghijklmnopqrs0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 20, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 20, 1, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S(""), 0, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S(""), 0, 1, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 0, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 0, 1, S("abcdefghijklmnopqrs1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 0, 2, S("abcdefghijklmnopqrs12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 0, 4, S("abcdefghijklmnopqrs1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 0, 5, S("abcdefghijklmnopqrs12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 0, 6, S("abcdefghijklmnopqrs12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 1, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 1, 1, S("abcdefghijklmnopqrs2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 1, 2, S("abcdefghijklmnopqrs23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 1, 3, S("abcdefghijklmnopqrs234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 1, 4, S("abcdefghijklmnopqrs2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 1, 5, S("abcdefghijklmnopqrs2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 2, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 2, 1, S("abcdefghijklmnopqrs3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 2, 2, S("abcdefghijklmnopqrs34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 2, 3, S("abcdefghijklmnopqrs345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 2, 4, S("abcdefghijklmnopqrs345"))));
    return 0;
}

template <class S>
static int test52()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 4, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 4, 1, S("abcdefghijklmnopqrs5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 4, 2, S("abcdefghijklmnopqrs5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 5, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 5, 1, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 0, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 0, 1, S("abcdefghijklmnopqrs1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 0, 5, S("abcdefghijklmnopqrs12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 0, 9, S("abcdefghijklmnopqrs123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 0, 10, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 0, 11, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 1, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 1, 1, S("abcdefghijklmnopqrs2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 1, 4, S("abcdefghijklmnopqrs2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 1, 8, S("abcdefghijklmnopqrs23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 1, 9, S("abcdefghijklmnopqrs234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 1, 10, S("abcdefghijklmnopqrs234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 5, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 5, 1, S("abcdefghijklmnopqrs6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 5, 2, S("abcdefghijklmnopqrs67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 5, 4, S("abcdefghijklmnopqrs6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 5, 5, S("abcdefghijklmnopqrs67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 5, 6, S("abcdefghijklmnopqrs67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 9, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 9, 1, S("abcdefghijklmnopqrs0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 9, 2, S("abcdefghijklmnopqrs0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 10, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 10, 1, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 0, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 0, 1, S("abcdefghijklmnopqrs1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 0, 10, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 0, 19, S("abcdefghijklmnopqrs1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 0, 20, S("abcdefghijklmnopqrs12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 0, 21, S("abcdefghijklmnopqrs12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 1, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 1, 1, S("abcdefghijklmnopqrs2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 1, 9, S("abcdefghijklmnopqrs234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 1, 18, S("abcdefghijklmnopqrs234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 1, 19, S("abcdefghijklmnopqrs2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 1, 20, S("abcdefghijklmnopqrs2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 10, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 10, 1, S("abcdefghijklmnopqrs1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 10, 5, S("abcdefghijklmnopqrs12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 10, 9, S("abcdefghijklmnopqrs123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 10, 10, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 10, 11, S("abcdefghijklmnopqrs1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 19, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 19, 1, S("abcdefghijklmnopqrs0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 19, 2, S("abcdefghijklmnopqrs0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 20, 0, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 20, 1, S("abcdefghijklmnopqrs"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 19, 2, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S(""), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S(""), 0, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 0, 1, S("abcdefghijklmnopqrst1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 0, 2, S("abcdefghijklmnopqrst12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 0, 4, S("abcdefghijklmnopqrst1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 0, 5, S("abcdefghijklmnopqrst12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 0, 6, S("abcdefghijklmnopqrst12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 1, 1, S("abcdefghijklmnopqrst2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 1, 2, S("abcdefghijklmnopqrst23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 1, 3, S("abcdefghijklmnopqrst234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 1, 4, S("abcdefghijklmnopqrst2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 1, 5, S("abcdefghijklmnopqrst2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 2, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 2, 1, S("abcdefghijklmnopqrst3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 2, 2, S("abcdefghijklmnopqrst34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 2, 3, S("abcdefghijklmnopqrst345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 2, 4, S("abcdefghijklmnopqrst345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 4, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 4, 1, S("abcdefghijklmnopqrst5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 4, 2, S("abcdefghijklmnopqrst5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 5, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 0, 1, S("abcdefghijklmnopqrst1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 0, 5, S("abcdefghijklmnopqrst12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 0, 9, S("abcdefghijklmnopqrst123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 0, 10, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 0, 11, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 1, 1, S("abcdefghijklmnopqrst2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 1, 4, S("abcdefghijklmnopqrst2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 1, 8, S("abcdefghijklmnopqrst23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 1, 9, S("abcdefghijklmnopqrst234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 1, 10, S("abcdefghijklmnopqrst234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 5, 1, S("abcdefghijklmnopqrst6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 5, 2, S("abcdefghijklmnopqrst67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 5, 4, S("abcdefghijklmnopqrst6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 5, 5, S("abcdefghijklmnopqrst67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 5, 6, S("abcdefghijklmnopqrst67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 9, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 9, 1, S("abcdefghijklmnopqrst0"))));
    return 0;
}

template <class S>
static int test53()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 9, 2, S("abcdefghijklmnopqrst0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 10, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 0, 1, S("abcdefghijklmnopqrst1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 0, 10, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 0, 19, S("abcdefghijklmnopqrst1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 0, 20, S("abcdefghijklmnopqrst12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 0, 21, S("abcdefghijklmnopqrst12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 1, 1, S("abcdefghijklmnopqrst2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 1, 9, S("abcdefghijklmnopqrst234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 1, 18, S("abcdefghijklmnopqrst234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 1, 19, S("abcdefghijklmnopqrst2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 1, 20, S("abcdefghijklmnopqrst2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 10, 1, S("abcdefghijklmnopqrst1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 10, 5, S("abcdefghijklmnopqrst12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 10, 9, S("abcdefghijklmnopqrst123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 10, 10, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 10, 11, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 19, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 19, 1, S("abcdefghijklmnopqrst0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 19, 2, S("abcdefghijklmnopqrst0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 20, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 20, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S(""), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S(""), 0, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 0, 1, S("abcdefghijklmnopqrst1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 0, 2, S("abcdefghijklmnopqrst12"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 0, 4, S("abcdefghijklmnopqrst1234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 0, 5, S("abcdefghijklmnopqrst12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 0, 6, S("abcdefghijklmnopqrst12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 1, 1, S("abcdefghijklmnopqrst2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 1, 2, S("abcdefghijklmnopqrst23"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 1, 3, S("abcdefghijklmnopqrst234"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 1, 4, S("abcdefghijklmnopqrst2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 1, 5, S("abcdefghijklmnopqrst2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 2, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 2, 1, S("abcdefghijklmnopqrst3"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 2, 2, S("abcdefghijklmnopqrst34"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 2, 3, S("abcdefghijklmnopqrst345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 2, 4, S("abcdefghijklmnopqrst345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 4, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 4, 1, S("abcdefghijklmnopqrst5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 4, 2, S("abcdefghijklmnopqrst5"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 5, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 0, 1, S("abcdefghijklmnopqrst1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 0, 5, S("abcdefghijklmnopqrst12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 0, 9, S("abcdefghijklmnopqrst123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 0, 10, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 0, 11, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 1, 1, S("abcdefghijklmnopqrst2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 1, 4, S("abcdefghijklmnopqrst2345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 1, 8, S("abcdefghijklmnopqrst23456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 1, 9, S("abcdefghijklmnopqrst234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 1, 10, S("abcdefghijklmnopqrst234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 5, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 5, 1, S("abcdefghijklmnopqrst6"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 5, 2, S("abcdefghijklmnopqrst67"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 5, 4, S("abcdefghijklmnopqrst6789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 5, 5, S("abcdefghijklmnopqrst67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 5, 6, S("abcdefghijklmnopqrst67890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 9, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 9, 1, S("abcdefghijklmnopqrst0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 9, 2, S("abcdefghijklmnopqrst0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 10, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 0, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 0, 1, S("abcdefghijklmnopqrst1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 0, 10, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 0, 19, S("abcdefghijklmnopqrst1234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 0, 20, S("abcdefghijklmnopqrst12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 0, 21, S("abcdefghijklmnopqrst12345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 1, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 1, 1, S("abcdefghijklmnopqrst2"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 1, 9, S("abcdefghijklmnopqrst234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 1, 18, S("abcdefghijklmnopqrst234567890123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 1, 19, S("abcdefghijklmnopqrst2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 1, 20, S("abcdefghijklmnopqrst2345678901234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 10, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 10, 1, S("abcdefghijklmnopqrst1"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 10, 5, S("abcdefghijklmnopqrst12345"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 10, 9, S("abcdefghijklmnopqrst123456789"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 10, 10, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 10, 11, S("abcdefghijklmnopqrst1234567890"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 19, 0, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 19, 1, S("abcdefghijklmnopqrst0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 19, 2, S("abcdefghijklmnopqrst0"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 20, 0, S("abcdefghijklmnopqrst"))));
    return 0;
}

template <class S>
static int test54()
{
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 20, 1, S("abcdefghijklmnopqrst"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 20, 1, S("12345678901234567890"), 21, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S(""), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S(""), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S(""), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 0, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 0, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 0, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 0, 6, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 1, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 1, 3, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 1, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 1, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 2, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 2, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 2, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 2, 3, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 2, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 4, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 4, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 4, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 5, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 5, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345"), 6, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 0, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 0, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 0, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 0, 11, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 1, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 1, 8, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 1, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 1, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 5, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 5, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 5, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 5, 4, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 5, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 5, 6, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 9, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 9, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 9, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 10, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 10, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("1234567890"), 11, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 0, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 0, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 0, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 0, 19, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 0, 20, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 0, 21, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 1, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 1, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 1, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 1, 18, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 1, 19, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 1, 20, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 10, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 10, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 10, 5, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 10, 9, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 10, 10, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 10, 11, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 19, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 19, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 19, 2, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 20, 0, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 20, 1, S("can't happen"))));
    TC_ASSERT_FUNC((test(S("abcdefghijklmnopqrst"), 21, 0, S("12345678901234567890"), 21, 0, S("can't happen"))));
    return 0;
}

template <class S>
static int test55()
{
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 1, S("12345678901234567890"), 10, S("abcdefghi1234567890"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 1, S("12345678901234567890"), 19, S("abcdefghi0"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 1, S("12345678901234567890"), 20, S("abcdefghi"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 1, S("12345678901234567890"), 20, S("abcdefghi"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 1, S("12345678901234567890"), 21, S("can't happen"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 2, S(""), 0, S("abcdefghi"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 2, S(""), 1, S("can't happen"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 2, S("12345"), 0, S("abcdefghi12345"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 2, S("12345"), 1, S("abcdefghi2345"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 2, S("12345"), 2, S("abcdefghi345"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 2, S("12345"), 4, S("abcdefghi5"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 2, S("12345"), 5, S("abcdefghi"))));
    TC_ASSERT_FUNC((test_npos(S("abcdefghij"), 9, 2, S("12345"), 6, S("can't happen"))));
    return 0;
}

int tc_libcxx_strings_string_replace_size_size_string_size_size(void)
{
    {
    typedef std::string S;
    TC_ASSERT_FUNC((test0<S>()));
    TC_ASSERT_FUNC((test1<S>()));
    TC_ASSERT_FUNC((test2<S>()));
    TC_ASSERT_FUNC((test3<S>()));
    TC_ASSERT_FUNC((test4<S>()));
    TC_ASSERT_FUNC((test5<S>()));
    TC_ASSERT_FUNC((test6<S>()));
    TC_ASSERT_FUNC((test7<S>()));
    TC_ASSERT_FUNC((test8<S>()));
    TC_ASSERT_FUNC((test9<S>()));
    TC_ASSERT_FUNC((test10<S>()));
    TC_ASSERT_FUNC((test11<S>()));
    TC_ASSERT_FUNC((test12<S>()));
    TC_ASSERT_FUNC((test13<S>()));
    TC_ASSERT_FUNC((test14<S>()));
    TC_ASSERT_FUNC((test15<S>()));
    TC_ASSERT_FUNC((test16<S>()));
    TC_ASSERT_FUNC((test17<S>()));
    TC_ASSERT_FUNC((test18<S>()));
    TC_ASSERT_FUNC((test19<S>()));
    TC_ASSERT_FUNC((test20<S>()));
    TC_ASSERT_FUNC((test21<S>()));
    TC_ASSERT_FUNC((test22<S>()));
    TC_ASSERT_FUNC((test23<S>()));
    TC_ASSERT_FUNC((test24<S>()));
    TC_ASSERT_FUNC((test25<S>()));
    TC_ASSERT_FUNC((test26<S>()));
    TC_ASSERT_FUNC((test27<S>()));
    TC_ASSERT_FUNC((test28<S>()));
    TC_ASSERT_FUNC((test29<S>()));
    TC_ASSERT_FUNC((test30<S>()));
    TC_ASSERT_FUNC((test31<S>()));
    TC_ASSERT_FUNC((test32<S>()));
    TC_ASSERT_FUNC((test33<S>()));
    TC_ASSERT_FUNC((test34<S>()));
    TC_ASSERT_FUNC((test35<S>()));
    TC_ASSERT_FUNC((test36<S>()));
    TC_ASSERT_FUNC((test37<S>()));
    TC_ASSERT_FUNC((test38<S>()));
    TC_ASSERT_FUNC((test39<S>()));
    TC_ASSERT_FUNC((test40<S>()));
    TC_ASSERT_FUNC((test41<S>()));
    TC_ASSERT_FUNC((test42<S>()));
    TC_ASSERT_FUNC((test43<S>()));
    TC_ASSERT_FUNC((test44<S>()));
    TC_ASSERT_FUNC((test45<S>()));
    TC_ASSERT_FUNC((test46<S>()));
    TC_ASSERT_FUNC((test47<S>()));
    TC_ASSERT_FUNC((test48<S>()));
    TC_ASSERT_FUNC((test49<S>()));
    TC_ASSERT_FUNC((test50<S>()));
    TC_ASSERT_FUNC((test51<S>()));
    TC_ASSERT_FUNC((test52<S>()));
    TC_ASSERT_FUNC((test53<S>()));
    TC_ASSERT_FUNC((test54<S>()));
    TC_ASSERT_FUNC((test55<S>()));
    }
    TC_SUCCESS_RESULT();
    return 0;
}
